"""
The :py:mod:`vc2_conformance.pseudocode` module contains implementations of
many of the VC-2 pseudocode functions and associated data structures. In
particular, it includes everything which can be used as-is in the construction
of encoders, decoders and so on without any alterations.

The VC-2 pseudocode is augmented by a modest number of additional functions or
data structure fields. For instance, additional fields are added to
:py:class:`~vc2_conformance.pseudocode.state.State` which are used by
conformance checking routines and forward wavelet transform routines.


:py:mod:`vc2_conformance.pseudocode.arrays`
-------------------------------------------

.. automodule:: vc2_conformance.pseudocode.arrays
    :members:

:py:mod:`vc2_conformance.pseudocode.offsetting`
-----------------------------------------------

.. automodule:: vc2_conformance.pseudocode.offsetting
    :members:


:py:mod:`vc2_conformance.pseudocode.parse_code_functions`
---------------------------------------------------------

.. automodule:: vc2_conformance.pseudocode.parse_code_functions
    :members:


:py:mod:`vc2_conformance.pseudocode.picture_decoding`
-----------------------------------------------------

.. automodule:: vc2_conformance.pseudocode.picture_decoding
    :members:
    :exclude-members: SYNTHESIS_LIFTING_FUNCTION_TYPES

.. autodata:: vc2_conformance.pseudocode.picture_decoding.SYNTHESIS_LIFTING_FUNCTION_TYPES
    :annotation: = {...}


:py:mod:`vc2_conformance.pseudocode.picture_encoding`
-----------------------------------------------------

.. automodule:: vc2_conformance.pseudocode.picture_encoding
    :members:
    :exclude-members: ANALYSIS_LIFTING_FUNCTION_TYPES

.. autodata:: vc2_conformance.pseudocode.picture_encoding.ANALYSIS_LIFTING_FUNCTION_TYPES
    :annotation: = {...}


:py:mod:`vc2_conformance.pseudocode.quantization`
-------------------------------------------------

.. automodule:: vc2_conformance.pseudocode.quantization
    :members:


:py:mod:`vc2_conformance.pseudocode.slice_sizes`
------------------------------------------------

.. automodule:: vc2_conformance.pseudocode.slice_sizes
    :members:


:py:mod:`vc2_conformance.pseudocode.state`
------------------------------------------

.. automodule:: vc2_conformance.pseudocode.state
    :members:


:py:mod:`vc2_conformance.pseudocode.vc2_math`
---------------------------------------------

.. automodule:: vc2_conformance.pseudocode.vc2_math
    :members:


:py:mod:`vc2_conformance.pseudocode.video_parameters`
-----------------------------------------------------

.. automodule:: vc2_conformance.pseudocode.video_parameters
    :members:

:py:mod:`vc2_conformance.pseudocode.metadata`
---------------------------------------------

.. automodule:: vc2_conformance.pseudocode.metadata

"""

from vc2_conformance.pseudocode.arrays import *
from vc2_conformance.pseudocode.offsetting import *
from vc2_conformance.pseudocode.parse_code_functions import *
from vc2_conformance.pseudocode.picture_decoding import *
from vc2_conformance.pseudocode.picture_encoding import *
from vc2_conformance.pseudocode.quantization import *
from vc2_conformance.pseudocode.slice_sizes import *
from vc2_conformance.pseudocode.state import *
from vc2_conformance.pseudocode.vc2_math import *
from vc2_conformance.pseudocode.video_parameters import *

from vc2_conformance.pseudocode.metadata import *



"""
This module defines the rules which define which version number a sequence must
be labelled with according to (11.2.2).
"""

from vc2_data_tables import Profiles

from vc2_conformance.pseudocode.state import State
from vc2_conformance.pseudocode.parse_code_functions import is_fragment


MINIMUM_MAJOR_VERSION = 1


def preset_frame_rate_version_implication(index):
    if index > 11:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def preset_signal_range_version_implication(index):
    if index > 4:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def preset_color_spec_version_implication(index):
    if index > 4:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def preset_color_primaries_version_implication(index):
    if index > 3:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def preset_color_matrix_version_implication(index):
    if index > 3:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def preset_transfer_function_version_implication(index):
    if index > 3:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def wavelet_transform_version_implication(
    wavelet_index, wavelet_index_ho, dwt_depth_ho
):
    if dwt_depth_ho != 0:
        return 3
    elif wavelet_index != wavelet_index_ho:
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


def parse_code_version_implication(parse_code):
    if is_fragment(State(parse_code=parse_code)):
        return 3
    else:
        return MINIMUM_MAJOR_VERSION


# Errata: Previously major_version 2 was specified if a high quality picture
# appeared in a sequence. This, however, is insufficient as an empty sequence
# marked with the 'high quality' profile would not be supported by an older
# decoder so a new condition on the profile field has been added.
def profile_version_implication(profile):
    if profile == Profiles.high_quality:
        return 2
    else:
        return MINIMUM_MAJOR_VERSION



"""
The :py:mod:`vc2_conformance.pseudocode.metadata` module is used to record the
relationship between the functions in the :py:mod:`vc2_conformance` software
and the pseudocode functions in the VC-2 specification documents. This
information has two main uses:

1. To produce more helpful error messages which include cross-references to
   published specifications.
2. To enable automatic verification that the behaviour of this software exactly
   matches the published specifications (see :py:mod:`verification`).


Implementing pseudocode functions
---------------------------------

Functions in the codebase which implement a pseudocode function in the
specification must be labelled as such using the :py:func:`ref_pseudocode`
decorator::

    >>> from vc2_conformance.pseudocode.metadata import ref_pseudocode

    >>> @ref_pseudocode
    ... def parse_info(state):
    ...     '''(10.5.1) Read a parse_info header.'''
    ...     byte_align()
    ...     read_uint_lit(state, 4)
    ...     state["parse_code"] = read_uint_lit(state, 1)
    ...     state["next_parse_offset"] = read_uint_lit(state, 4)
    ...     state["previous_parse_offset"] = read_uint_lit(state, 4)

The :py:func:`ref_pseudocode` decorator will log the existence of the decorated
function, along with the reference to the spec at the start of its docstring.

By default, it is implied that a decorated function does *not* deviate from the
pseudocode -- a fact which is verified by the :py:mod:`verification` module in
the test suite. When a function does deviate (for example for use in
serialisation/deserialisation) or is not defined in the spec, the ``deviation``
argument should be provided. This is used by the :py:mod:`verification` logic
to determine how the function will be verified. See
:ref:`verification-deviation-types` for the allowed values. For example::

    >>> @ref_pseudocode(deviation="serdes")
    ... def parse_parameters(serdes, state):
    ...     '''(11.2.1)'''
    ...     state["major_version"] = serdes.uint("major_version")
    ...     state["minor_version"] = serdes.uint("minor_version")
    ...     state["profile"] = serdes.uint("profile")
    ...     state["level"] = serdes.uint("level")

.. autofunction:: ref_pseudocode


Accessing the metadata
----------------------

After all submodules of :py:mod:`vc2_conformance` have been loaded, the
:py:data:`pseudocode_derived_functions` list will be populated with
:py:class:`PseudocodeDerivedFunction` instances for every pseudocode derived
function annotated with :py:func:`ref_pseudocode`.

.. autoclass:: PseudocodeDerivedFunction
    :members:

.. autodata:: pseudocode_derived_functions
    :annotation: = [PseudocodeDerivedFunction, ...]


Pseudocode tracebacks
---------------------

The :py:func:`make_pseudocode_traceback` function may be used to convert a
Python traceback into a pseudocode function traceback.

.. autofunction:: make_pseudocode_traceback

"""

import re
import inspect

__all__ = [
    "DEFAULT_SPECIFICATION",
    "PseudocodeDerivedFunction",
    "pseudocode_derived_functions",
    "ref_pseudocode",
    "make_pseudocode_traceback",
]


DEFAULT_SPECIFICATION = "SMPTE ST 2042-1:2017"  # The main VC-2 specification
"""
If not otherwise specified, the specification section numbers refer to.
"""


class PseudocodeDerivedFunction(object):
    """
    A record of a pseudocode-derived function within this codebase.

    Parameters
    ==========
    function : function
        The Python function which was derived from a pseudocode function.
    deviation : str or None
        If this function deviates from the pseudocode in any way, indicates the
        nature of this deviation. None indicates that this function exactly
        matches the pseudocode.

        Automatic checks for pseudocode equivalence in the test suite (see
        :py:mod:`verification`) will use this value to determine how exactly
        this function must match the pseudocode for tests to pass. See
        :ref:`verification-deviation-types` for details of values this field
        may hold.
    document, section : str or None
        The name of the document and section defining this function. If None,
        will be extracted from the function docstring. Defaults to the main
        VC-2 specification. If not provided, and not found in the docstring, a
        :py:exc:`TypeError` will be thrown.

        When extracted from a docstring, a reference of one of the following
        forms must be used at the start of the docstring:

        * ``(1.2.3)``: Reference a section in the main VC-2 spec
        * ``(SMPTE ST 20402-2:2017: 1.2.3)``: Reference a section in another
          spec
    name : str or None
        The name of the pseudocode function this function implements. If None,
        the provided function's name will be used.
    """

    def __init__(
        self,
        function,
        deviation=None,
        document=None,
        section=None,
        name=None,
    ):
        # Auto-extract section/document
        if section is None or document is None:
            autodetected_document = None
            autodetected_section = None
            docstring = getattr(function, "__doc__", "")
            match = re.match(r"\s*\(([^\)]+)\)", docstring)
            if match:
                autodetected_document = DEFAULT_SPECIFICATION
                autodetected_section = match.group(1)

                # Also extract the document name, if present
                submatch = re.match(r"(.*)\s*:\s*([^:]+)$", autodetected_section)
                if submatch:
                    autodetected_document = submatch.group(1)
                    autodetected_section = submatch.group(2)

            if section is None:
                section = autodetected_section

            if document is None:
                document = autodetected_document

        if section is None or document is None:
            raise TypeError("No document or section reference could be found.")

        # Auto-extract function name
        if name is None:
            name = getattr(function, "__name__", None)

        self.function = function
        self.deviation = deviation
        self.document = document
        self.section = section
        self.name = name

    @property
    def citation(self):
        """
        A human-readable citation string for this pseudocode function.
        """
        return "{} ({}{})".format(
            self.name,
            (
                "{}: ".format(self.document)
                if self.document != DEFAULT_SPECIFICATION
                else ""
            ),
            self.section,
        )


pseudocode_derived_functions = []
r"""
The complete set of :py:class:`PseudocodeDerivedFunctions
<PseudocodeDerivedFunction>` in this code base, in no particular order.

.. warning::

    This array is only completely populated once every module of
    :py:mod:`vc2_conformance` has been loaded.
"""


def ref_pseudocode(*args, **kwargs):
    """
    Decorator for marking functions which are derived from the VC-2 pseudocode.

    Example usage::

        @ref_pseudocode
        def parse_info(state):
            ''''(10.5.1) Read a parse_info header.'''
            # ...

        @ref_pseudocode("10.5.1", deviation="alternative_implementation")
        def parse_info(state):
            '''Some alternative implementation of parse info.'''

    Parameters
    ==========
    deviation : str or None
        If this function deviates from the pseudocode in any way, indicates the
        nature of this deviation. None indicates that this function exactly
        matches the pseudocode.

        Automatic checks for pseudocode equivalence in the test suite (see
        :py:mod:`verification`) will use this value to determine how exactly
        this function must match the pseudocode for tests to pass. See
        :ref:`verification-deviation-types` for details of values this field
        may hold.
    document, section : str or None
        The name of the document and section defining this function. If None,
        will be extracted from the function docstring. Defaults to the main
        VC-2 specification. If not provided, and not found in the docstring, a
        :py:exc:`TypeError` will be thrown.

        When extracted from a docstring, a reference of one of the following
        forms must be used at the start of the docstring:

        * ``(1.2.3)``: Reference a section in the main VC-2 spec
        * ``(SMPTE ST 20402-2:2017: 1.2.3)``: Reference a section in another
          spec
    name : str or None
        The name of the pseudocode function this function implements. If None,
        the provided function's name will be used.
    """
    if len(args) >= 1 and callable(args[0]):
        pseudocode_derived_functions.append(PseudocodeDerivedFunction(*args, **kwargs))
        return args[0]
    else:

        def decorator(f):
            pseudocode_derived_functions.append(
                PseudocodeDerivedFunction(f, *args, **kwargs)
            )
            return f

        return decorator


def make_pseudocode_traceback(tb):
    """
    Given a :py:func:`traceback.extract_tb` generated traceback description,
    return a list of :py:class:`PseudocodeDerivedFunction` objects for the
    stack trace, most recently called last. Entries in the traceback which
    don't have a corresponding pseudocode-derived function are omitted.
    """
    calls = []

    for frame_summary in tb:
        filename = frame_summary[0]
        name = frame_summary[2]

        try:
            pseudocode_derived_function = next(
                iter(
                    filter(
                        lambda e: (
                            e.function.__name__ == name
                            and inspect.getsourcefile(e.function) == filename
                        ),
                        pseudocode_derived_functions,
                    )
                )
            )
            calls.append(pseudocode_derived_function)
        except StopIteration:
            # Not a pseudocode function, omit this from the traceback
            pass

    return calls



"""
This module contains the wavelet synthesis filters and associated functions
defined in the pseudocode of the VC-2 standard (15).

See also :py:mod:`vc2_conformance.pseudocode.picture_encoding`.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_data_tables import LIFTING_FILTERS, LiftingFilterTypes

from vc2_conformance.pseudocode.vc2_math import clip

from vc2_conformance.pseudocode.arrays import (
    new_array,
    width,
    height,
    row,
    column,
    delete_rows_after,
    delete_columns_after,
)

__all__ = [
    "inverse_wavelet_transform",
    "idwt",
    "h_synthesis",
    "vh_synthesis",
    "oned_synthesis",
    "filter_bit_shift",
    "lift1",
    "lift2",
    "lift3",
    "lift4",
    "SYNTHESIS_LIFTING_FUNCTION_TYPES",
    "idwt_pad_removal",
    "offset_picture",
    "offset_component",
    "clip_picture",
    "clip_component",
]


@ref_pseudocode
def picture_decode(state):
    """(15.2)"""
    state["current_picture"] = {}
    state["current_picture"]["pic_num"] = state["picture_number"]
    inverse_wavelet_transform(state)
    clip_picture(state, state["current_picture"])
    offset_picture(state, state["current_picture"])

    # In this concrete implementation, the 'output_picture' function is
    # provided via the _output_picture_callback comes via State.

    ### output_picture(
    ###     state["current_picture"],
    ###     state["video_parameters"],
    ###     state["picture_coding_mode"],
    ### )

    ## Begin not in spec
    if "_output_picture_callback" in state:
        state["_output_picture_callback"](
            state["current_picture"],
            state["video_parameters"],
            state["picture_coding_mode"],
        )
    ## End not in spec


@ref_pseudocode
def inverse_wavelet_transform(state):
    """(15.3)"""
    state["current_picture"]["Y"] = idwt(state, state["y_transform"])
    state["current_picture"]["C1"] = idwt(state, state["c1_transform"])
    state["current_picture"]["C2"] = idwt(state, state["c2_transform"])
    for c in ["Y", "C1", "C2"]:
        idwt_pad_removal(state, state["current_picture"][c], c)


################################################################################
# Inverse (Synthesis) Discrete Wavelet Transform
################################################################################


@ref_pseudocode
def idwt(state, coeff_data):
    """
    (15.4.1)

    Inverse Discrete Wavelet Transform.

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least the following:

        * ``wavelet_index``
        * ``wavelet_index_ho``
        * ``dwt_depth``
        * ``dwt_depth_ho``

    coeff_data : {level: {orientation: [[coeff, ...], ...], ...}, ...}
        The complete (power-of-two dimensioned) transform coefficient data.

    Returns
    =======
    picture : [[pixel_value, ...], ...]
        The synthesized picture.
    """
    if state["dwt_depth_ho"] == 0:
        DC_band = coeff_data[0]["LL"]
    else:
        DC_band = coeff_data[0]["L"]
    for n in range(1, state["dwt_depth_ho"] + 1):
        new_DC_band = h_synthesis(state, DC_band, coeff_data[n]["H"])
        DC_band = new_DC_band
    for n in range(
        state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
    ):
        new_DC_band = vh_synthesis(
            state,
            DC_band,
            coeff_data[n]["HL"],
            coeff_data[n]["LH"],
            coeff_data[n]["HH"],
        )
        DC_band = new_DC_band
    return DC_band


@ref_pseudocode
def h_synthesis(state, L_data, H_data):
    """(15.4.2) Horizontal-only synthesis."""
    synth = new_array(height(L_data), 2 * width(L_data))

    # Interleave transform data (as expected by synthesis routine)
    for y in range(height(synth)):
        for x in range(width(synth) // 2):
            synth[y][2 * x] = L_data[y][x]
            synth[y][(2 * x) + 1] = H_data[y][x]

    # Synthesis
    for y in range(height(synth)):
        oned_synthesis(row(synth, y), state["wavelet_index_ho"])

    # Bit shift, if required
    shift = filter_bit_shift(state)
    if shift > 0:
        for y in range(height(synth)):
            for x in range(width(synth)):
                synth[y][x] = (synth[y][x] + (1 << (shift - 1))) >> shift

    return synth


@ref_pseudocode
def vh_synthesis(state, LL_data, HL_data, LH_data, HH_data):
    """(15.4.3) Interleaved vertical and horizontal synthesis."""
    synth = new_array(2 * height(LL_data), 2 * width(LL_data))

    # Interleave transform data (as expected by synthesis routine)
    for y in range(height(synth) // 2):
        for x in range(width(synth) // 2):
            synth[2 * y][2 * x] = LL_data[y][x]
            synth[2 * y][2 * x + 1] = HL_data[y][x]
            synth[2 * y + 1][2 * x] = LH_data[y][x]
            synth[2 * y + 1][2 * x + 1] = HH_data[y][x]

    # Synthesis
    for x in range(width(synth)):
        oned_synthesis(column(synth, x), state["wavelet_index"])
    for y in range(height(synth)):
        oned_synthesis(row(synth, y), state["wavelet_index_ho"])

    # Bit shift, if required
    shift = filter_bit_shift(state)
    if shift > 0:
        for y in range(height(synth)):
            for x in range(width(synth)):
                synth[y][x] = (synth[y][x] + (1 << (shift - 1))) >> shift

    return synth


@ref_pseudocode(deviation="inferred_implementation")
def oned_synthesis(A, filter_index):
    """(15.4.4.1) and (15.4.4.3). Acts in-place on 'A'"""
    filter_params = LIFTING_FILTERS[filter_index]

    for stage in filter_params.stages:
        lift_fn = SYNTHESIS_LIFTING_FUNCTION_TYPES[stage.lift_type]
        lift_fn(A, stage.L, stage.D, stage.taps, stage.S)


@ref_pseudocode(deviation="inferred_implementation")
def filter_bit_shift(state):
    """(15.4.2) Return the bit shift for the current horizontal-only filter."""
    filter_params = LIFTING_FILTERS[state["wavelet_index_ho"]]
    return filter_params.filter_bit_shift


@ref_pseudocode
def lift1(A, L, D, taps, S):
    """(15.4.4.1) Update even, add odd."""
    for n in range(len(A) // 2):
        sum = 0
        for i in range(D, L + D):
            pos = 2 * (n + i) - 1
            pos = min(pos, len(A) - 1)
            pos = max(pos, 1)
            sum += taps[i - D] * A[pos]
        if S > 0:
            sum += 1 << (S - 1)
        A[2 * n] += sum >> S


@ref_pseudocode
def lift2(A, L, D, taps, S):
    """(15.4.4.1) Update even, subtract odd."""
    for n in range(len(A) // 2):
        sum = 0
        for i in range(D, L + D):
            pos = 2 * (n + i) - 1
            pos = min(pos, len(A) - 1)
            pos = max(pos, 1)
            sum += taps[i - D] * A[pos]
        if S > 0:
            sum += 1 << (S - 1)
        A[2 * n] -= sum >> S


@ref_pseudocode
def lift3(A, L, D, taps, S):
    """(15.4.4.1) Update odd, add even."""
    for n in range(len(A) // 2):
        sum = 0
        for i in range(D, L + D):
            pos = 2 * (n + i)
            pos = min(pos, len(A) - 2)
            pos = max(pos, 0)
            sum += taps[i - D] * A[pos]
        if S > 0:
            sum += 1 << (S - 1)
        A[2 * n + 1] += sum >> S


@ref_pseudocode
def lift4(A, L, D, taps, S):
    """(15.4.4.1) Update odd, subtract even."""
    for n in range(len(A) // 2):
        sum = 0
        for i in range(D, L + D):
            pos = 2 * (n + i)
            pos = min(pos, len(A) - 2)
            pos = max(pos, 0)
            sum += taps[i - D] * A[pos]
        if S > 0:
            sum += 1 << (S - 1)
        A[2 * n + 1] -= sum >> S


SYNTHESIS_LIFTING_FUNCTION_TYPES = {
    LiftingFilterTypes(1): lift1,
    LiftingFilterTypes(2): lift2,
    LiftingFilterTypes(3): lift3,
    LiftingFilterTypes(4): lift4,
}
"""
Lookup from lifting function ID to function implementation for implementing
wavelet synthesis.
"""


################################################################################
# Padding removal
################################################################################


@ref_pseudocode
def idwt_pad_removal(state, pic, c):
    """(15.4.5)"""
    if c == "Y":
        width = state["luma_width"]
        height = state["luma_height"]
    elif (c == "C1") or (c == "C2"):
        width = state["color_diff_width"]
        height = state["color_diff_height"]

    delete_rows_after(pic, height)
    delete_columns_after(pic, width)


################################################################################
# Pixel value offsetting and clipping
################################################################################


@ref_pseudocode
def offset_picture(state, current_picture):
    """(15.5) Remove picture value offsets (added during encoding).

    Parameters
    ==========
    state : :py:class:`vc2_conformance.pseudocode.state.State`
        Where ``luma_depth`` and ``color_diff_depth`` are defined.
    current_picture : {comp: [[pixel_value, ...], ...], ...}
        Will be mutated in-place.
    """
    for c in ["Y", "C1", "C2"]:
        offset_component(state, current_picture[c], c)


@ref_pseudocode
def offset_component(state, comp_data, c):
    """(15.5) Remove picture value offsets from a single component."""
    for y in range(height(comp_data)):
        for x in range(width(comp_data)):
            if c == "Y":
                comp_data[y][x] += 2 ** (state["luma_depth"] - 1)
            else:
                comp_data[y][x] += 2 ** (state["color_diff_depth"] - 1)


@ref_pseudocode
def clip_picture(state, current_picture):
    """(15.5)"""
    for c in ["Y", "C1", "C2"]:
        clip_component(state, current_picture[c], c)


@ref_pseudocode
def clip_component(state, comp_data, c):
    """(15.5)"""
    for y in range(height(comp_data)):
        for x in range(width(comp_data)):
            if c == "Y":
                comp_data[y][x] = clip(
                    comp_data[y][x],
                    -(2 ** (state["luma_depth"] - 1)),
                    2 ** (state["luma_depth"] - 1) - 1,
                )
            else:
                comp_data[y][x] = clip(
                    comp_data[y][x],
                    -(2 ** (state["color_diff_depth"] - 1)),
                    2 ** (state["color_diff_depth"] - 1) - 1,
                )



r"""
.. _vc2-test-case-generator:

``vc2-test-case-generator``
===========================

This application generates test cases for VC-2 encoders and decoders given a
set of codec features to target.

* For a guide to the codec features CSV file format see :ref:`codec-features`
* For a guide to encoder test cases see :ref:`encoder-test-cases`
* For a guide to decoder test cases see :ref:`decoder-test-cases`

Usage
-----

A codec features CSV file should be created (as described in
:ref:`guide-generating-test-cases`). This can then be processed by
``vc2-test-case-generator`` to generate test cases for the codecs supporting
the specific codec features described.

Test case generation is performed as follows::

    $ vc2-test-case-generator path/to/codec_features.csv

The test cases are written to a ``test_cases`` in the current working
directory. The ``--output`` argument can be given to choose an alternative
location. By default, if any existing test cases are present in the output
directory, the test case generator will not run. The ``--force`` argument can
be added to run the test case generator anyway.

If only a subset of test cases are required, the following arguments can also
be used:

* ``--encoder-only``: Only generate test cases for encoders
* ``--decoder-only``: Only generate test cases for decoders
* ``--codec <regexp>``: Only generate test cases for codec feature sets whose
  names match a supplied pattern.


Parallel execution
------------------

The test case generator can take many hours to run. To speed up execution,
test cases can be generated in parallel on multi-core machines. To do this the
``--parallel`` argument is added to ``vc2-test-case-generator``. Rather
than performing test case generation, this command outputs a series of
newline-separated commands which can be executed in parallel to perform test
case generation. These can then be executed, for example using `GNU Parallel
<https://www.gnu.org/software/parallel/>`_::

    $ # Write test case generation commands to 'commands.txt'
    $ vc2-test-case-generator path/to/codec_features.csv --parallel > commands.txt

    $ # Run test case generation in parallel using GNU Parallel
    $ parallel -a commands.txt


Static wavelet filter analyses
------------------------------

Some test cases require a precomputed mathematical analysis of the wavelet
transform used.

The VC-2 conformance software includes suitable analyses for all wavelet
transforms for which a default quantisation matrix is defined in the VC-2
specification. To support additional wavelet transform combinations and depths,
additional analyses must be generated as described in
:ref:`generating-static-analyses`.

If the ``VC2_BIT_WIDTHS_BUNDLE`` environment variable is defined, the test case
generator will attempt to read wavelet filter analyses from the file it
specifies. This must be a filter analysis bundle file created by
``vc2-bundle``.


Arguments
---------

The complete set of arguments can be listed using ``--help``

.. program-output:: vc2-test-case-generator --help

"""

import os

import sys

import re

import logging

import json

from functools import partial

from collections import OrderedDict

from io import BytesIO

from argparse import ArgumentParser

from vc2_conformance import __version__

from vc2_conformance.codec_features import (
    read_codec_features_csv,
    InvalidCodecFeaturesError,
)

from vc2_conformance.encoder import UnsatisfiableCodecFeaturesError

from vc2_conformance.py2x_compat import (
    get_terminal_size,
    makedirs,
    FileType,
)

from vc2_conformance.string_utils import wrap_paragraphs

from vc2_conformance import file_format

from vc2_conformance.color_conversion import sanity_check_video_parameters

from vc2_conformance.test_cases.decoder import static_gray

from vc2_conformance.bitstream import autofill_and_serialise_stream

from vc2_conformance.pseudocode.state import State

from vc2_conformance.decoder import (
    init_io,
    parse_stream,
    ConformanceError,
)

from vc2_conformance.test_cases import (
    ENCODER_TEST_CASE_GENERATOR_REGISTRY,
    DECODER_TEST_CASE_GENERATOR_REGISTRY,
)

from vc2_conformance.scripts.vc2_test_case_generator.worker import create_command


def regex(regex):
    try:
        return re.compile("^{}$".format(regex))
    except Exception as e:
        raise ValueError(str(e))


def parse_args(*args, **kwargs):
    """
    Parse a set of command line arguments. Returns a :py:mod:`argparse`
    ``args`` object.
    """
    parser = ArgumentParser(
        description="""
            Generate test inputs for VC-2 encoder and decoder implementations.
        """
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__),
    )

    parser.add_argument(
        "codec_configurations",
        # NB: utf-8-sig encoding ignores any Unicode BOM appearing at the start
        # of the file sometimes inserted by Excel (and other Microsoft
        # products).
        type=FileType("r", encoding="utf-8-sig"),
        help="""
            CSV file containing the set of codec configurations to generate
            test cases for.
        """,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="""
            Show additional status information during execution.
        """,
    )

    parser.add_argument(
        "--parallel",
        "-p",
        action="store_true",
        default=False,
        help="""
            If given, don't generate test cases but instead produce a series of
            commands on stdout which can be executed in parallel to generate
            the test cases.
        """,
    )

    parser.add_argument(
        "--output",
        "-o",
        default="./test_cases/",
        help="""
            Directory name to write test cases to. Will be created if it does
            not exist. Defaults to '%(default)s'.
        """,
    )

    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        default=False,
        help="""
            Force the test case generator to run even if the output directory
            is not empty.
        """,
    )

    encoder_decoder_only_group = parser.add_mutually_exclusive_group()

    encoder_decoder_only_group.add_argument(
        "--encoder-only",
        "-e",
        action="store_true",
        default=False,
        help="""
            If set, only generate test cases for VC-2 encoders.
        """,
    )

    encoder_decoder_only_group.add_argument(
        "--decoder-only",
        "-d",
        action="store_true",
        default=False,
        help="""
            If set, only generate test cases for VC-2 decoders.
        """,
    )

    parser.add_argument(
        "--codecs",
        "-c",
        type=regex,
        default=regex(r".*"),
        metavar="REGEX",
        help="""
            If given, a regular expression which selects which codec
            configurations to generate test cases for. If not given, test cases
            will be generated for all codec configurations.
        """,
    )

    return parser.parse_args(*args, **kwargs)


def load_codec_features(csv, name_re):
    """
    Load a set of codec features from the specified CSV, filtering just those
    matching the provided compiled regular expression.

    Returns an :py:class:`collections.OrderedDict` of
    :py:class:`vc2_conformance.codec_features.CodecFeatures` objects. Calls
    :py:func:`sys.exit` and prints an error to stderr upon any problems (e.g.
    invalid or empty file).
    """
    try:
        codec_feature_sets = read_codec_features_csv(csv)
    except InvalidCodecFeaturesError as e:
        sys.stderr.write("Error: Invalid codec features: {}\n".format(e))
        sys.exit(1)

    # Filter to just the specified codec features columns
    codec_feature_sets = OrderedDict(
        (name, codec_features)
        for name, codec_features in codec_feature_sets.items()
        if name_re.match(name)
    )
    if len(codec_feature_sets) == 0:
        sys.stderr.write("Error: No matching codec feature sets found.\n")
        sys.exit(2)
    logging.info(
        "Loaded %s matching codec feature sets: %s",
        len(codec_feature_sets),
        ", ".join(codec_feature_sets),
    )

    return codec_feature_sets


def check_output_directories_empty(
    output_root_dir,
    codec_feature_sets,
    check_encoder_tests,
    check_decoder_tests,
):
    """
    Check all output directories are empty (or don't yet exist).  Prints an
    error to stderr and calls :py:func:`sys.exit` if a non empty directory is
    encountered.
    """
    for codec_features_name in codec_feature_sets:
        base_path = os.path.join(
            output_root_dir,
            codec_features_name,
        )

        paths = []
        if check_encoder_tests:
            paths.append(os.path.join(base_path, "encoder"))
        if check_decoder_tests:
            paths.append(os.path.join(base_path, "decoder"))

        for path in paths:
            try:
                if len(os.listdir(path)) > 0:
                    sys.stderr.write(
                        (
                            "Error: Output directory {!r} is not empty. "
                            "(Use --force/-f to disable this check)\n"
                        ).format(path)
                    )
                    sys.exit(3)
            except OSError:
                # Directory does not exist, good!
                pass


def check_codec_features_valid(codec_feature_sets):
    """
    Verify that the codec features requested don't themselves violate the spec
    (e.g. violate a level constraint). This is done by generating then
    validating a bitstream containing a single mid-gray frame.

    Prints an error to stderr and calls :py:func:`sys.exit` if a problem is
    encountered.
    """
    logging.info("Checking codec feature sets are valid...")
    for name, codec_features in codec_feature_sets.items():
        logging.info("Checking %r...", name)
        f = BytesIO()

        # Sanity-check the color format (since this won't raise a validation
        # error but will result in a useless gamut being available).
        sanity = sanity_check_video_parameters(codec_features["video_parameters"])
        if not sanity:
            logging.warning(
                "Color specification for codec configuration %r be malformed: %s",
                name,
                sanity.explain(),
            )

        # Generate a minimal bitstream
        try:
            autofill_and_serialise_stream(f, static_gray(codec_features))
        except UnsatisfiableCodecFeaturesError as e:
            sys.stderr.write(
                "Error: Codec configuration {!r} is invalid:\n".format(name)
            )
            terminal_width = get_terminal_size()[0]
            sys.stderr.write(wrap_paragraphs(e.explain(), terminal_width))
            sys.stderr.write("\n")
            sys.exit(4)
        f.seek(0)

        # Validate it meets the spec
        state = State()
        init_io(state, f)
        try:
            parse_stream(state)
        except ConformanceError as e:
            sys.stderr.write(
                "Error: Codec configuration {!r} is invalid:\n".format(name)
            )
            terminal_width = get_terminal_size()[0]
            sys.stderr.write(wrap_paragraphs(e.explain(), terminal_width))
            sys.stderr.write("\n")
            sys.exit(4)


def output_encoder_test_case(output_dir, codec_features, test_case):
    """
    Write an encoder test case to disk.

    Parameters
    ==========
    output_dir : str
        Output directory to write test cases to.
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    test_case : :py:class:`~vc2_conformance.test_cases.TestCase`
    """
    # Write raw pictures
    for i, picture in enumerate(test_case.value.pictures):
        picture_directory = os.path.join(
            output_dir,
            test_case.name,
        )
        makedirs(picture_directory, exist_ok=True)
        file_format.write(
            picture,
            test_case.value.video_parameters,
            test_case.value.picture_coding_mode,
            os.path.join(
                picture_directory,
                "picture_{}.raw".format(i),
            ),
        )
    # Write metadata
    if test_case.metadata is not None:
        with open(
            os.path.join(
                output_dir,
                "{}_metadata.json".format(test_case.name),
            ),
            "w",
        ) as f:
            json.dump(test_case.metadata, f)

    logging.info(
        "Generated encoder test case %s for %s",
        test_case.name,
        codec_features["name"],
    )


def output_encoder_test_cases(output_dir, codec_features, generator_function):
    for test_case in generator_function():
        output_encoder_test_case(output_dir, codec_features, test_case)


def output_decoder_test_case(output_dir, codec_features, test_case):
    """
    Write a decoder test case to disk.

    Parameters
    ==========
    output_dir : str
        Output directory to write test cases to.
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    test_case : :py:class:`~vc2_conformance.test_cases.TestCase`
    """
    # Serialise bitstream
    bitstream_filename = os.path.join(
        output_dir,
        "{}.vc2".format(test_case.name),
    )
    with open(bitstream_filename, "wb") as f:
        autofill_and_serialise_stream(f, test_case.value)

    # Decode model answer
    model_answer_directory = os.path.join(
        output_dir,
        "{}_expected".format(test_case.name),
    )
    makedirs(model_answer_directory, exist_ok=True)
    with open(bitstream_filename, "rb") as f:
        index = [0]

        def output_picture(picture, video_parameters, picture_coding_mode):
            file_format.write(
                picture,
                video_parameters,
                picture_coding_mode,
                os.path.join(
                    model_answer_directory,
                    "picture_{}.raw".format(index[0]),
                ),
            )
            index[0] += 1

        state = State(_output_picture_callback=output_picture)
        init_io(state, f)
        parse_stream(state)

    # Write metadata
    if test_case.metadata is not None:
        with open(
            os.path.join(
                output_dir,
                "{}_metadata.json".format(test_case.name),
            ),
            "w",
        ) as f:
            json.dump(test_case.metadata, f)

    logging.info(
        "Generated decoder test case %s for %s",
        test_case.name,
        codec_features["name"],
    )


def output_decoder_test_cases(output_dir, codec_features, generator_function):
    makedirs(output_dir, exist_ok=True)
    for test_case in generator_function():
        output_decoder_test_case(output_dir, codec_features, test_case)


def set_log_level_and_call(log_level, fn, *args, **kwargs):
    logging.basicConfig(level=log_level)
    return fn(*args, **kwargs)


def main(*args, **kwargs):
    args = parse_args(*args, **kwargs)

    log_level = logging.WARNING
    if args.verbose >= 2:
        log_level = logging.DEBUG
    elif args.verbose >= 1:
        log_level = logging.INFO
    logging.basicConfig(level=log_level)

    codec_feature_sets = load_codec_features(
        args.codec_configurations,
        args.codecs,
    )

    if not args.force:
        check_output_directories_empty(
            args.output,
            codec_feature_sets,
            not args.decoder_only,
            not args.encoder_only,
        )

    check_codec_features_valid(codec_feature_sets)

    to_call = []

    for name, codec_features in codec_feature_sets.items():
        if not args.decoder_only:
            output_dir = os.path.join(
                args.output,
                name,
                "encoder",
            )
            for (
                generator_function
            ) in ENCODER_TEST_CASE_GENERATOR_REGISTRY.iter_independent_generators(
                codec_features,
            ):
                to_call.append(
                    partial(
                        set_log_level_and_call,
                        log_level,
                        output_encoder_test_cases,
                        output_dir,
                        codec_features,
                        generator_function,
                    )
                )

        if not args.encoder_only:
            output_dir = os.path.join(
                args.output,
                name,
                "decoder",
            )
            for (
                generator_function
            ) in DECODER_TEST_CASE_GENERATOR_REGISTRY.iter_independent_generators(
                codec_features,
            ):
                to_call.append(
                    partial(
                        set_log_level_and_call,
                        log_level,
                        output_decoder_test_cases,
                        output_dir,
                        codec_features,
                        generator_function,
                    )
                )

    if args.parallel:
        for fn in to_call:
            print(create_command(fn))
    else:
        for fn in to_call:
            fn()

    return 0



from vc2_conformance.test_cases.decoder.sequence_header import *
from vc2_conformance.test_cases.decoder.padding import *
from vc2_conformance.test_cases.decoder.pictures import *
from vc2_conformance.test_cases.decoder.picture_numbers import *
from vc2_conformance.test_cases.decoder.signal_range import *
from vc2_conformance.test_cases.decoder.real_pictures import *
from vc2_conformance.test_cases.decoder.extended_transform_parameters import *
from vc2_conformance.test_cases.decoder.slice_prefix_bytes import *
from vc2_conformance.test_cases.decoder.absent_next_parse_offset import *
from vc2_conformance.test_cases.decoder.slice_size_scaler import *
from vc2_conformance.test_cases.decoder.custom_quantization_matrix import *
from vc2_conformance.test_cases.decoder.concatenated_sequences import *
from vc2_conformance.test_cases.decoder.lossless_quantization import *



"""
The :py:mod:`vc2_conformance.symbol_re` module contains a regular expression
matching system for sequences of data unit types in VC-2 bitstreams.

This module has two main applications: checking the order of data units during
validation and generating valid sequences during bitstream generation.

During bitstream validation (see :py:mod:`vc2_conformance.decoder`) the
validator must check that sequences contain the right pattern of data unit
types. For example, some levels might require bitstreams to include a sequence
header between each picture while others may require fragmented and
non-fragmented picture types are not used in the same stream. These rules are
described by regular expressions which are evaluated by this module.

During bitstream generation, the encoder (see
:py:mod:`vc2_conformance.encoder`) must produce sequences conforming to the
above patterns. To facilitate this, this module can also *generate* sequence
orderings which fulfil the rules described by regular expressions.


Regular expressions of symbols
------------------------------

Unlike string-matching regular expression libraries (e.g. :py:mod:`re`) which
match sequences of characters, this module is designed to match sequences of
'symbols' where each symbol is just a string like ``"sequence_header"`` or
``"high_quality_picture"``.

As an example, we might write the following regular expression to describe the
rule 'all sequences must start with a sequence header and end with an
end of sequence data unit'::

    sequence_header .* end_of_sequence

In the regular expression syntax used by this module, whitespace is ignored but
otherwise follows the typical form for regular expressions. For example, ``.``
is a wildcard which matches any symbol and ``*`` means 'zero or more
occurrences of the previous pattern'.

This regular expression would match the following sequence of symbols (data
unit type names)::

    "sequence_header"
    "high_quality_picture"
    "sequence_header"
    "high_quality_picture"
    "end_of_sequence"

But would not match the following sequence (because it does not start with a
sequence header)::

    "high_quality_picture"
    "sequence_header"
    "high_quality_picture"
    "end_of_sequence"


Matching VC-2 sequences
-----------------------

The :py:class:`Matcher` class implements a regular expression matching state
machine, taking the regular expression to be matched as its argument.  By
convention, we use the parse code names defined in
:py:class:`~vc2_data_tables.ParseCodes` as symbols to represent each type of
data unit in a VC-2 sequence.

.. enum-value-table:: vc2_data_tables.ParseCodes
    :value-heading: Data unit parse code
    :name-heading: Symbol

In this example below we'll create a :py:class:`Matcher` which checks if a
sequence consists of alternating sequence headers and high quality picture data
units::

    >>> from vc2_conformance.symbol_re import Matcher

    >>> m = Matcher("(sequence_header high_quality_picture)* end_of_sequence")

This :py:class:`Matcher` instance is then used to check if a sequence matches
the required pattern by feeding it symbols one at a time via the
:py:meth:`~Matcher.match_symbol` method. This returns True so long as the
sequence matches to expression::

    >>> m.match_symbol("sequence_header")
    True
    >>> m.match_symbol("high_quality_picture")
    True
    >>> m.match_symbol("sequence_header")
    True
    >>> m.match_symbol("high_quality_picture")
    True

When we reach the end of the sequence, the :py:meth:`~Matcher.is_complete`
method will check to see if the regular expression has completely matched the
sequence (i.e. it isn't expecting any other data units)::

    >>> # Missing the required end_of_sequence!
    >>> m.is_complete()
    False

    >>> # Now complete
    >>> m.match_symbol("end_of_sequence")
    True
    >>> m.is_complete()
    True

When a non-matching sequence is encountered, the
:py:meth:`Matcher.valid_next_symbols` method may be used to enumerate which
symbols the regular expression matcher was expecting. For example::

    >>> # NB: Each Matcher can only be used once so we must create a new one
    >>> # for this new sequence!
    >>> m = Matcher("(sequence_header high_quality_picture)* end_of_sequence")

    >>> m.match_symbol("sequence_header")
    True
    >>> m.match_symbol("high_quality_picture")
    True
    >>> m.match_symbol("high_quality_picture")  # Not allowed!
    False
    >>> m.valid_next_symbols()
    {"sequence_header", "end_of_sequence"}


Generating VC-2 sequences
-------------------------

This module provides the :py:func:`make_matching_sequence` function which can
generate minimal sequences matching a set of regular expressions. For example,
say we wish to generate a sequence containing three pictures matching the
regular expression we used in our previous example::

    >>> from vc2_conformance.symbol_re import make_matching_sequence

    >>> from pprint import pprint
    >>> pprint(make_matching_sequence(
    ...     ["high_quality_picture"]*3,
    ...     "(sequence_header high_quality_picture)* end_of_sequence",
    ... ))
    ['sequence_header',
     'high_quality_picture',
     'sequence_header',
     'high_quality_picture',
     'sequence_header',
     'high_quality_picture',
     'end_of_sequence']

Here, the sequence headers and the final end of sequence data units have been
added automatically.

API
---

.. autoclass:: Matcher
    :members:

.. autofunction:: make_matching_sequence

.. autodata:: WILDCARD

.. autodata:: END_OF_SEQUENCE

.. autoexception:: SymbolRegexSyntaxError

.. autoexception:: ImpossibleSequenceError


Internals
---------

Beyond the parts exposed by the public API above, this module internally is
built on top of two main parts:

* A parser which parses the regular expression syntax accepted by
  :py:class:`Matcher` into an Abstract Syntax Tree (AST)
* A Non-deterministic Finite-state Automaton (NFA) representation which is
  constructed from the AST using `Thompson's constructions
  <https://en.wikipedia.org/wiki/Thompson%27s_construction>`_.

The parser is broken into two stages: a simple tokenizer/lexer
(:py:func:`tokenize_regex`) and a recursive descent parser
(:py:func:`parse_expression`). A utility function combining these steps is
provided by :py:func:`parse_regex`.

.. autofunction:: tokenize_regex

.. autofunction:: parse_expression

.. autofunction:: parse_regex

The parser outputs an AST constructed from the following
elements:

.. autoclass:: Symbol

.. autoclass:: Star

.. autoclass:: Concatenation

.. autoclass:: Union

An NFA can be constructed from an AST using the :py:meth:`NFA.from_ast` class
method.

.. autoclass:: NFA
    :members:

.. autoclass:: NFANode
    :members:
"""


import re

from copy import deepcopy

from collections import defaultdict, namedtuple, deque


__all__ = [
    "Matcher",
    "SymbolRegexSyntaxError",
    "WILDCARD",
    "END_OF_SEQUENCE",
    "ImpossibleSequenceError",
    "make_matching_sequence",
]


class SymbolRegexSyntaxError(Exception):
    """
    Thrown when a regular expression string is provided which could not be
    parsed.
    """


TOKEN_REGEX = re.compile(
    r"(?P<string>[\w]+)|"
    r"(?P<wildcard>[.])|"
    r"(?P<end_of_sequence>[$])|"
    r"(?P<modifier>[?*+])|"
    r"(?P<bar>[|])|"
    r"(?P<parenthesis>[()])"
)
"""
A regular expression which matches a single token in the regular expression
syntax.
"""


def tokenize_regex(regex_string):
    """
    A generator which tokenizes a sequence regular expression specification
    into (token_type, token_value, offset) 3-tuples.

    Token types are:

    * ``"string"`` (value is the string)
    * ``"modifier"`` (value is one of ``?*+``)
    * ``"wildcard"`` (value is ``.``)
    * ``"end_of_sequence"`` (value is ``$``)
    * ``"bar"`` (value is ``|``)
    * ``"parenthesis"`` (value is one of ``()``)

    Throws a :py:exc:`SymbolRegexSyntaxError` if an invalid character is
    encountered.
    """
    # Remove newline characters which have special behaviour in Python re
    # library
    regex_string = regex_string.replace("\n", " ").replace("\r", " ")

    offset = 0
    while True:
        # Skip whitespace
        ws_match = re.match(r"^\s*", regex_string)
        regex_string = regex_string[ws_match.end() :]
        offset += ws_match.end()

        # Special case: End of string
        if not regex_string:
            break

        # Process token
        t_match = TOKEN_REGEX.match(regex_string)
        if not t_match:
            raise SymbolRegexSyntaxError(
                "Unexpected text at position {}".format(offset)
            )
        [(token_type, token_value)] = [
            (t, v) for t, v in t_match.groupdict().items() if v is not None
        ]
        yield (token_type, token_value, offset)
        regex_string = regex_string[t_match.end() :]
        offset += t_match.end()


Symbol = namedtuple("Symbol", "symbol")
"""Leaf AST node for a symbol."""

Star = namedtuple("Star", "expr")
"""AST node for a Kleene Star pattern (``*``)."""

Concatenation = namedtuple("Concatenation", "a,b")
"""AST node for a concatenation of two expressions."""

Union = namedtuple("Union", "a,b")
"""AST node for a union (``|``) of two expressions."""


WILDCARD = "."
"""A constant representing a wildcard match."""

END_OF_SEQUENCE = ""
"""A constant representing the end-of-sequence."""


def parse_expression(tokens):
    """
    A recursive-descent parser which parses an Abstract Syntax Tree (AST) from
    the regex specification.

    The parsed tokens will be removed from the provided token list. Tokens are
    consumed right-to-left making implementing tight binding of modifiers (i.e.
    '?', '*' and '+') easy.

    This function will return as soon as it runs out of tokens or reaches an
    unmatched opening parenthesis.

    Returns an AST node: one of :py:class:`Symbol`, :py:class:`Star`,
    :py:class:`Concatenation`, :py:class:`Union` or ``None``.
    """
    ast = None
    modifier = None

    while tokens and tokens[-1][0:2] != ("parenthesis", "("):
        # Special case: reached a modifier: just set a flag
        if tokens[-1][0] == "modifier":
            if modifier is not None:
                raise SymbolRegexSyntaxError(
                    "Multiple modifiers at position {}".format(tokens[-1][2])
                )
            modifier = tokens.pop(-1)[1]
            continue

        # Special case: reached a union
        if tokens[-1][0] == "bar":
            if modifier is not None:
                raise SymbolRegexSyntaxError(
                    "Modifier before '|' at position {}".format(tokens[-1][2])
                )
            tokens.pop(-1)
            ast = Union(parse_expression(tokens), ast)
            continue

        # Grab the next expression from the token list
        if tokens[-1][0:2] == ("parenthesis", ")"):
            tokens.pop(-1)
            next_ast = parse_expression(tokens)
            if not tokens:
                raise SymbolRegexSyntaxError("Unmatched parentheses")
            tokens.pop(-1)
        elif tokens[-1][0] == "string":
            next_ast = Symbol(tokens.pop(-1)[1])
        elif tokens[-1][0] == "wildcard":
            tokens.pop(-1)
            next_ast = Symbol(WILDCARD)
        elif tokens[-1][0] == "end_of_sequence":
            tokens.pop(-1)
            next_ast = Symbol(END_OF_SEQUENCE)

        # Apply the modifier (as required)
        if modifier == "*":
            next_ast = Star(next_ast)
        elif modifier == "+":
            next_ast = Concatenation(next_ast, Star(next_ast))
        elif modifier == "?":
            next_ast = Union(next_ast, None)
        modifier = None

        # Add the expression to the AST
        if ast is None:
            ast = next_ast
        else:
            ast = Concatenation(next_ast, ast)

    if modifier is not None:
        if tokens:
            raise SymbolRegexSyntaxError(
                "Modifier before '(' at position {}".format(tokens[-1][2])
            )
        else:
            raise SymbolRegexSyntaxError("Modifier at start of expression")

    return ast


def parse_regex(regex_string):
    """
    Parse a sequence regular expression specification into an Abstract Syntax
    Tree (AST) consisting of None (empty), :py:class:`Symbol`, :py:class:`Star`
    :py:class:`Concatenation` and :py:class:`Union` objects.
    """
    tokens = list(tokenize_regex(regex_string))

    ast = parse_expression(tokens)
    if tokens:
        raise SymbolRegexSyntaxError("Unmatched parentheses")

    return ast


class NFANode(object):
    """
    A node (i.e. state) in a Non-deterministic Finite-state Automaton (NFA).

    Attributes
    ==========
    transitions : {symbol: set([:py:class:`NFANode`, ...]), ...}
        The transition rules from this node.

        Empty transitions are listed under the symbol ``None`` and are always
        bidirectional.
    """

    def __init__(self):
        self.transitions = defaultdict(set)

    def add_transition(self, dest_node, symbol=None):
        """
        Add a transition rule from this node to the specified destination.

        If no symbols are specified, a (bidirectional) empty transition between
        the two nodes will be added.
        """
        if symbol is None:
            # Empty transitions should be bidirectional
            self.transitions[symbol].add(dest_node)
            dest_node.transitions[symbol].add(self)
        else:
            self.transitions[symbol].add(dest_node)

    def equivalent_nodes(self):
        """
        Iterate over the set of :py:class:`NFANode` nodes connected to this one
        by only empty transitions (includes this node).
        """
        visited = set([self])
        to_visit = [self]
        while to_visit:
            node = to_visit.pop()
            yield node

            for other in node.transitions.get(None, []):
                if other not in visited:
                    to_visit.append(other)
                    visited.add(other)

    def follow(self, symbol):
        r"""
        Iterate over the :py:class:`NFANodes <NFANode>` reachable from this
        node following the given symbol.
        """
        visited = set()
        for node in self.equivalent_nodes():
            for neighbour in node.transitions.get(symbol, []):
                if neighbour not in visited:
                    yield neighbour
                    visited.add(neighbour)


class NFA(object):
    """
    A Non-deterministic Finite-state Automaton (NFA) with a labelled 'start'
    and 'final' state.

    Attributes
    ==========
    start : :py:class:`NFANode`
    final : :py:class:`NFANode`
    """

    def __init__(self, start=None, final=None):
        self.start = start or NFANode()
        self.final = final or NFANode()

    @classmethod
    def from_ast(cls, ast):
        """
        Convert a regular expression AST node into a new :py:class:`NFA` object
        using `Thompson's constructions
        <https://en.wikipedia.org/wiki/Thompson%27s_construction>`_.
        """
        if ast is None:
            node = NFANode()
            return cls(node, node)
        elif isinstance(ast, Symbol):
            nfa = cls()
            nfa.start.add_transition(nfa.final, ast.symbol)

            return nfa
        elif isinstance(ast, Concatenation):
            nfa_a = cls.from_ast(ast.a)
            nfa_b = cls.from_ast(ast.b)

            nfa_a.final.add_transition(nfa_b.start)

            return cls(nfa_a.start, nfa_b.final)
        elif isinstance(ast, Symbol):
            nfa = cls()
            nfa.start.add_transition(nfa.final, ast.symbol)
            return nfa
        elif isinstance(ast, Union):
            nfa = cls()

            nfa_a = cls.from_ast(ast.a)
            nfa_b = cls.from_ast(ast.b)

            nfa.start.add_transition(nfa_a.start)
            nfa.start.add_transition(nfa_b.start)

            nfa_a.final.add_transition(nfa.final)
            nfa_b.final.add_transition(nfa.final)

            return nfa
        elif isinstance(ast, Star):
            nfa = cls()

            sub_nfa = cls.from_ast(ast.expr)

            nfa.start.add_transition(nfa.final)
            nfa.start.add_transition(sub_nfa.start)

            sub_nfa.final.add_transition(sub_nfa.start)
            sub_nfa.final.add_transition(nfa.final)

            return nfa


class Matcher(object):
    """
    Test whether a sequence of symbols (alpha-numeric strings with underscores,
    e.g. ``"foo"`` or ``"bar_123"``) conforms to a pattern described by a
    regular expression.

    :py:meth:`match_symbol` should be called for each symbol in the sequence.
    If ``False`` is returned, the sequence does not match the specified regular
    expression. :py:meth:`valid_next_symbols` may be used to list what symbols
    *would* have been allowed at this stage of the sequence. Once the entire
    sequence has been passed to :py:meth:`match_symbol`, :py:meth:`is_complete`
    should be used to check that a complete pattern has been matched.

    .. note::

        Each instance of :py:class:`Matcher` can only be used to match a single
        sequence. To match another sequence a new :py:class:`Matcher` must be
        created.

    Parameters
    ==========
    pattern : str
        The regular expression describing the pattern this :py:class:`Matcher`
        should match.

        Regular expressions may be specified with a syntax similar (but
        different to) that used by common string-matching regular expression
        libraries.

        * An alpha-numeric-with-underscore expression matches a single instance
          of the specified symbol. For example ``foo_123`` will match the
          symbol ``"foo_123"``.
        * A dot (``.``) will match any (singular) symbol.
        * A dollar (``$``) will match only at the end of the sequence.
        * Two expressions (separated by any amount of whitespace) will match
          the first expression followed by the second. For example ``. foo``
          will match a sequence ``"anything"``, ``"foo"``.
        * Two expressions separated by a bar (``|``) will match either the
          first expression or the second expression.
        * A question mark (``?``) suffix to an expression will match zero or
          one instances of that expression. For example ``foo?`` will match an
          empty sequence or a single ``"foo"``.
        * An asterisk (``*``) suffix to an expression will match zero or more
          instances of that expression. For example ``foo*`` will match an
          empty sequence, a sequence of a single ``"foo"`` symbol or a sequence
          of many ``"foo"`` symbols.
        * A plus (``+``) suffix to an expression will match one or more
          instances of that expression. For example ``foo+`` will match a
          sequence of a single ``"foo"`` symbol or a sequence of many ``"foo"``
          symbols.
        * Parentheses (``(`` and ``)``) may be used to group expressions
          together into a single logical expression.

        The expression suffixes bind tightly to their left-hand expression.
        Beyond this, consider operator precedence undefined: be explicit to
        help readability!
    """

    def __init__(self, pattern):
        # This object directly executes the NFA of the provided regular
        # expression. The 'cur_states' set holds the set of states we've
        # reached in the NFA.
        self.nfa = NFA.from_ast(parse_regex(pattern))
        self.cur_states = set([self.nfa.start])

    def match_symbol(self, symbol):
        """
        Attempt to match the next symbol in the sequence.

        Returns True if the symbol matched and False otherwise.

        If no symbol was matched, the state machine will not be advanced (i.e.
        you can try again with a different symbol as if nothing happened).
        """
        new_states = set()
        for nfa in self.cur_states:
            new_states.update(nfa.follow(symbol))
            new_states.update(nfa.follow(WILDCARD))

        if not new_states:
            return False

        self.cur_states = new_states
        return True

    def is_complete(self):
        """
        Is it valid for the sequence to terminate at this point?
        """
        for nfa in self.cur_states:
            # Is final state in NFA
            if self.nfa.final in list(nfa.equivalent_nodes()):
                return True
            # ...or is explicit end-of-sequence marker
            if list(nfa.follow(END_OF_SEQUENCE)):
                return True

        return False

    def valid_next_symbols(self):
        """
        Return the :py:class:`set` of valid next symbols in the sequence.

        If a wildcard is allowed, :py:data:`WILDCARD` will be returned as one
        of the symbols in addition to any concretely allowed symbols.

        If it is valid for the sequence to end at this point,
        :py:data:`END_OF_SEQUENCE` will be in the returned set.
        """
        valid_symbols = set()
        for nfa in self.cur_states:
            for equivalent_nfa in nfa.equivalent_nodes():
                for symbol in equivalent_nfa.transitions:
                    if symbol is not None:
                        valid_symbols.add(symbol)

        # If we're allowed to end the string here, also include
        # END_OF_SEQUENCE.
        if self.is_complete():
            valid_symbols.add(END_OF_SEQUENCE)

        return valid_symbols


class ImpossibleSequenceError(Exception):
    """
    Thrown when :py:func:`make_matching_sequence` is unable to find a suitable
    sequence of symbols.
    """

    pass


def make_matching_sequence(initial_sequence, *patterns, **kwargs):
    """
    Given a sequence of symbols, returns a new sequence based on this which
    matches the supplied set of patterns. The new sequence will be a copy of
    the supplied sequence with additional symbols inserted where necessary.

    Parameters
    ==========
    initial_sequence : [symbol, ...]
        The minimal set of entries which must be included in the sequence, in
        the order they are required to appear.
    patterns : str
        A series of one or more regular expression specifications (as accepted
        by :py:class:`Matcher`) which the generated sequence must
        simultaneously satisfy.
    depth_limit : int
        Keyword-only argument specifying the maximum number of consecutive
        symbols to try inserting before giving up on finding a matching
        sequence. Defaults to 4.
    symbol_priority : [symbol, ...]
        Keyword-only argument. If supplied, orders possible symbols from most
        to least preferable. Though this function will always return a sequence
        of the shortest possible length, when several equal-length sequences
        would be valid, this argument may be used to influence which is
        returned.  Any symbols not appearing in the list
        will be given the lowest priority, and sorted in alphabetical order.
        If this argument is not supplied (or is empty), whenever 'wildcard'
        value (``.``) is required by a regular expression, the
        :py:data:`WILDCARD` sentinel will be inserted into the output sequence
        rather than a concrete symbol.

    Returns
    =======
    matching_sequence : [symbol, ...]
        The shortest sequence of symbols which satisfies all of the supplied
        regular expression patterns.  This will contain a superset of the
        sequence in ``initial_sequence``, that is one where additional symbols
        may have been inserted but none are removed or reordered.

    Raises
    ======
    ImpossibleSequenceError
        Thrown if no sequence of symbols could be found which matches all of
        the supplied patterns.
    """
    # NB: Python 2.x doesn't directly support keyword-only oarguments
    depth_limit = kwargs.pop("depth_limit", 3)
    symbol_priority = kwargs.pop("symbol_priority", [])
    if kwargs:
        raise TypeError(
            "find_minimal_sequence() got unuexpected keyword argument(s) {}".format(
                ", ".join(map(repr, kwargs)),
            )
        )

    # This function performs a breadth-first search of the pattern space.

    # Queue of candidates to try
    #     (symbols_so_far, symbols_remaining, matchers, this_depth_limit)
    # Where:
    # * symbols_so_far is the symbols of the generated sequence so far
    # * symbols_remaining is the list of symbols from initial_sequence still to
    #   included
    # * matchers is a list of Matcher objects which have matched the
    #   symbols_so_far
    # * this_depth_limit is an integer giving the number of search levels
    #   remaining before giving up.
    initial_matchers = [Matcher(pattern) for pattern in patterns]
    queue = deque([([], initial_sequence, initial_matchers, depth_limit)])

    while queue:
        (
            symbols_so_far,
            symbols_remaining,
            matchers,
            this_depth_limit,
        ) = queue.popleft()

        # Try and match the next required symbool
        if len(symbols_remaining) == 0:
            if all(m.is_complete() for m in matchers):
                # No more symbols to match and found a suitable matching
                # sequence! We're done.
                return symbols_so_far
        else:
            if all(
                symbols_remaining[0] in m.valid_next_symbols()
                or WILDCARD in m.valid_next_symbols()
                for m in matchers
            ):
                # The next symbol is matched by all matchers, move on!
                new_matchers = deepcopy(matchers)
                for m in new_matchers:
                    m.match_symbol(symbols_remaining[0])
                queue.append(
                    (
                        symbols_so_far + [symbols_remaining[0]],
                        symbols_remaining[1:],
                        new_matchers,
                        depth_limit,  # NB: Reset depth limit when a match is found
                    )
                )
                continue

        # If we reach this point the current symbol in the provided sequence
        # was not matched by all of the matchers. We must now try inserting
        # some other symbol into the sequence and see if it lets us get any
        # further.

        if this_depth_limit <= 0:
            # Depth limit reached, give up on this branch of the search
            continue

        # Find the set of candidate symbols which would be accepted by all of
        # the matchers
        candidate_next_symbols = set([WILDCARD])
        for matcher in matchers:
            symbols = matcher.valid_next_symbols()
            symbols.discard(END_OF_SEQUENCE)
            if WILDCARD in symbols and WILDCARD in candidate_next_symbols:
                candidate_next_symbols.update(symbols)
            elif WILDCARD in candidate_next_symbols:
                candidate_next_symbols = symbols
            elif WILDCARD in symbols:
                pass
            else:
                candidate_next_symbols.intersection_update(symbols)

        if len(candidate_next_symbols) == 0:
            # Reached a dead-end (no symbol fits all patterns), give up on this
            # branch of the search
            continue

        # Descend the search into each of the potential next steps. We try
        # candidates in the order indicated by the symbol_priority argument.
        if WILDCARD in candidate_next_symbols and len(symbol_priority) > 0:
            # Substitute wildcard for concrete symbols if possible
            candidate_next_symbols.remove(WILDCARD)
            candidate_next_symbols.update(symbol_priority)
        candidate_symbols = sorted(
            candidate_next_symbols,
            key=lambda sym: (
                (symbol_priority.index(sym), None)
                if sym in symbol_priority
                else (len(symbol_priority), sym)
            ),
        )
        for candidate_symbol in candidate_symbols:
            new_matchers = deepcopy(matchers)
            for m in new_matchers:
                m.match_symbol(candidate_symbol)
            queue.append(
                (
                    symbols_so_far + [candidate_symbol],
                    symbols_remaining,
                    new_matchers,
                    this_depth_limit - 1,
                )
            )
            continue

    raise ImpossibleSequenceError()



"""
The :py:mod:`vc2_conformance.encoder.sequence` module provides routines for
constructing complete VC-2 sequences.

Principally, this module implements the :py:func:`make_sequence` function which
produces :py:class:`vc2_conformance.bitstream.Sequence` objects containing
pictures compressed according to the required codec specifications. This is the
main entry point to the encoder.

.. autofunction:: make_sequence

"""

from functools import partial

from itertools import repeat

from vc2_data_tables import ParseCodes

from vc2_conformance.py2x_compat import zip

from vc2_conformance.level_constraints import LEVEL_SEQUENCE_RESTRICTIONS

from vc2_conformance.bitstream import (
    Sequence,
    DataUnit,
    ParseInfo,
    Padding,
    AuxiliaryData,
)

from vc2_conformance.symbol_re import make_matching_sequence, ImpossibleSequenceError

from vc2_conformance.encoder.exceptions import IncompatibleLevelAndDataUnitError

from vc2_conformance.encoder.sequence_header import make_sequence_header_data_unit

from vc2_conformance.encoder.pictures import make_picture_data_units


__all__ = [
    "make_sequence",
]


def make_end_of_sequence_data_unit():
    return DataUnit(
        parse_info=ParseInfo(parse_code=ParseCodes.end_of_sequence),
    )


def make_auxiliary_data_unit():
    return DataUnit(
        parse_info=ParseInfo(parse_code=ParseCodes.auxiliary_data),
        auxiliary_data=AuxiliaryData(),
    )


def make_padding_data_unit():
    return DataUnit(
        parse_info=ParseInfo(parse_code=ParseCodes.padding_data),
        padding=Padding(),
    )


def make_sequence(codec_features, pictures, *data_unit_patterns, **kwargs):
    """
    Generate a complete VC-2 bitstream based on the provided set of codec
    features and containing compressed versions of the specified set of
    pictures.

    This function also takes a small number of additional parameters which
    override certain encoder behaviours as may be required by some test case
    generators.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    pictures : [{"Y": [[s, ...], ...], "C1": ..., "C2": ..., "pic_num": int}, ...]
        The pictures to be encoded in the bitstream. If ``pic_num`` is omitted,
        ``picture_number`` fields will also be omitted in the output (and left
        for, e.g.
        :py:func:`vc2_conformance.bitstream.autofill_and_serialise_stream` to
        assign). See :py:mod:`vc2_conformance.encoder.pictures` for details of
        the picture compression process.
    *data_unit_patterns : str
        Force the generated sequence of data units to match a specified regular
        expression. For example, ``"(. padding_data)+ end_of_sequence"`` will
        force a padding data unit to be inserted between each data unit. See
        the :py:mod:`vc2_conformance.symbol_re` module for details of the
        regular expression format.

        A sequence of data units matching all specified patterns while meeting
        the requirements of the VC-2 standard will be generated. If this is not
        possible,
        :py:exc:`vc2_conformance.encoder.exceptions.IncompatibleLevelAndDataUnitError`
        will be raised.
    minimum_qindex : int or [int, ...]
        Keyword-only argument. Default 0. Specifies the minimum quantization
        index to be used for all picture slices. If a list is provided,
        specifies the minimum quantization index separately for each picture.

        This option may be used by test cases where a particular (very high)
        quantization index must be used. Note that the encoder may still use
        larger quantization indices if a set of transform coefficients still do
        not fit into a slice so the caller must check that this has not
        occurred.

        Must be 0 for lossless coding modes.
    minimum_slice_size_scaler : int
        Keyword-only argument. Default 1. Specifies the minimum slice size
        scaler to use.

        For almost all sensible coding modes, the ``slice_size_scaler`` can be
        set to '1' -- and this encoder will do so if possible. To facilitate
        the production of test cases verifying higher values are supported,
        this option may be used to pick a larger value. The encoder may still
        use larger ``slice_size_scaler`` values if this is necessary, however.

        Only has an effect on high quality profile coding modes, will be
        ignored for the low delay profile modes.

    Returns
    =======
    sequence : :py:class:`vc2_conformance.bitstream.Sequence`
        The VC-2 bitstream sequence. This may be serialised by encapsulating
        it in a :py:class:`vc2_conformance.bitstream.Stream` and serialising it
        with
        :py:func:`~vc2_conformance.bitstream.vc2_autofill.autofill_and_serialise_stream`.

    Raises
    ======
    UnsatisfiableCodecFeaturesError
        Raised if a sequence could not be generated according to the
        requirements given.
    """
    minimum_qindices = kwargs.pop("minimum_qindex", 0)
    minimum_slice_size_scaler = kwargs.pop("minimum_slice_size_scaler", 1)
    assert not kwargs, "Unexpected arguments: {}".format(kwargs)

    if not isinstance(minimum_qindices, list):
        minimum_qindices = repeat(minimum_qindices)

    pictures_only_sequence = Sequence(data_units=[])
    for picture, minimum_qindex in zip(pictures, minimum_qindices):
        pictures_only_sequence["data_units"].extend(
            make_picture_data_units(
                codec_features,
                picture,
                minimum_qindex,
                minimum_slice_size_scaler,
            )
        )

    # Fill in all other required bitstream data units
    picture_only_data_unit_names = [
        data_unit["parse_info"]["parse_code"].name
        for data_unit in pictures_only_sequence["data_units"]
    ]
    try:
        required_data_unit_names = make_matching_sequence(
            picture_only_data_unit_names,
            # (10.4.1) Sequences start with a sequence header and end with an end
            # of sequence data unit
            "sequence_header .* end_of_sequence",
            # Certain levels may provide additional constraints
            LEVEL_SEQUENCE_RESTRICTIONS[
                codec_features["level"]
            ].sequence_restriction_regex,
            *data_unit_patterns,
            symbol_priority=["padding_data", "sequence_header"]
        )
    except ImpossibleSequenceError:
        raise IncompatibleLevelAndDataUnitError(codec_features)

    # Functions for producing the necessary data units
    #
    # {data_unit_name: fn() -> DataUnit, ...}
    data_unit_makers = {
        "sequence_header": partial(make_sequence_header_data_unit, codec_features),
        "end_of_sequence": make_end_of_sequence_data_unit,
        "auxiliary_data": make_auxiliary_data_unit,
        "padding_data": make_padding_data_unit,
    }
    if pictures_only_sequence["data_units"]:
        picture_parse_code = pictures_only_sequence["data_units"][0]["parse_info"][
            "parse_code"
        ]
        data_unit_makers[picture_parse_code.name] = partial(
            pictures_only_sequence["data_units"].pop,
            0,
        )

    return Sequence(
        data_units=[
            data_unit_makers[data_unit_name]()
            for data_unit_name in required_data_unit_names
        ]
    )



from vc2_data_tables import Profiles

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import decoder_test_case_generator

from vc2_conformance.encoder import make_sequence

from vc2_conformance.encoder.pictures import get_safe_lossy_hq_slice_size_scaler

from vc2_conformance.constraint_table import allowed_values_for

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.codec_features import codec_features_to_trivial_level_constraints

from vc2_conformance.picture_generators import mid_gray

from vc2_conformance.test_cases.decoder.common import iter_slices_in_sequence


@decoder_test_case_generator
def slice_size_scaler(codec_features):
    """
    **Tests that the 'slice_size_scaler' field is correctly handled.**

    This test case generates a sequence which sets slice_size_scaler value
    (13.5.4) 1 larger than it otherwise would be.

    This test case is only generated for the high quality profile, and levels
    which permit a slice size scaler value greater than 1.
    """
    # Skip if not high quality profile
    if codec_features["profile"] != Profiles.high_quality:
        return None

    # Pick a minimum slice size scaler which is larger than the slice size
    # scaler which would otherwise be used
    if codec_features["lossless"]:
        # We're just going to code mid-gray frames which compress to 0 bytes so
        # slice size scaler = 1 is always sufficient.
        minimum_slice_size_scaler = 2
    else:
        minimum_slice_size_scaler = (
            get_safe_lossy_hq_slice_size_scaler(
                codec_features["picture_bytes"],
                codec_features["slices_x"] * codec_features["slices_y"],
            )
            + 1
        )

    # Skip if level prohibits non-1 slice size scaler
    if minimum_slice_size_scaler not in allowed_values_for(
        LEVEL_CONSTRAINTS,
        "slice_size_scaler",
        codec_features_to_trivial_level_constraints(codec_features),
    ):
        return None

    sequence = make_sequence(
        codec_features,
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
        minimum_slice_size_scaler=minimum_slice_size_scaler,
    )

    # Force lossless coding modes to use a non-zero number of bytes for each
    # slice's coefficients (so that slice_size_scaler actually has to be used).
    if codec_features["lossless"]:
        for _state, _sx, _sy, hq_slice in iter_slices_in_sequence(
            codec_features, sequence
        ):
            assert hq_slice["slice_c2_length"] == 0
            hq_slice["slice_c2_length"] = 1

    return Stream(sequences=[sequence])



r"""
The :py:mod:`vc2_conformance.bitstream.vc2_fixeddicts` module contains
:py:mod:`~vc2_conformance.fixeddict` definitions for holding VC-2 bitstream
values in a hierarchy which strongly mimics the bitstream structure. These
names are re-exported in the :py:mod:`vc2_conformance.bitstream` module for
convenience. See :ref:`bitstream-fixeddicts` for a listing.

It also provides the following metadata structures:

.. autodata:: vc2_fixeddict_nesting
    :annotation:

.. autodata:: vc2_default_values
    :annotation:
"""

from bitarray import bitarray

from vc2_conformance.fixeddict import fixeddict, Entry

from vc2_conformance.string_formatters import (
    Hex,
    Bool,
    Bits,
    Bytes,
    List,
    Object,
    MultilineList,
)

from vc2_data_tables import (
    PARSE_INFO_PREFIX,
    ParseCodes,
    PictureCodingModes,
    BaseVideoFormats,
    Profiles,
    Levels,
    ColorDifferenceSamplingFormats,
    SourceSamplingModes,
    PresetFrameRates,
    PresetPixelAspectRatios,
    PresetSignalRanges,
    PresetColorSpecs,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
    WaveletFilters,
)


__all__ = [
    "vc2_fixeddict_nesting",
    "vc2_default_values",
    "ParseInfo",
    "SequenceHeader",
    "ParseParameters",
    "SourceParameters",
    "FrameSize",
    "ColorDiffSamplingFormat",
    "ScanFormat",
    "FrameRate",
    "PixelAspectRatio",
    "CleanArea",
    "SignalRange",
    "ColorSpec",
    "ColorPrimaries",
    "ColorMatrix",
    "TransferFunction",
    "AuxiliaryData",
    "Padding",
    "TransformParameters",
    "ExtendedTransformParameters",
    "SliceParameters",
    "QuantMatrix",
    "PictureHeader",
    "TransformData",
    "WaveletTransform",
    "PictureParse",
    "LDSlice",
    "HQSlice",
    "FragmentHeader",
    "FragmentData",
    "FragmentParse",
    "DataUnit",
    "Sequence",
    "Stream",
]


vc2_fixeddict_nesting = {}
"""
A lookup ``{fixeddict_type: [fixeddict_type, ...], ...}``.

This lookup enumerates the fixeddicts which may be directly contained by the
fixed dictionary types in this module.

This hierarchical information is used by user-facing tools to allow (e.g.)
recursive selection of a particular dictionary to display.
"""


vc2_default_values = {}
"""
A lookup ``{fixeddict_type: {key: default_value, ...}, ...}``

For each fixeddict type below, provides a sensible default value for each key.
The defaults are generally chosen to produce a minimal, but valid, bitstream.

Where a particular fixeddict entry is a list, the value listed in this lookup
should be treated as the default value to use for list entries.

.. warning::

    For default values containing a :py:class:`bitarray.bitarray` or any other
    mutable type, users must take care to copy the default value before
    mutating it.
"""


################################################################################
# parse_info header
################################################################################

ParseInfo = fixeddict(
    "ParseInfo",
    Entry(
        "padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Byte alignment padding bits.",
    ),
    Entry(
        "parse_info_prefix",
        friendly_formatter=(
            lambda prefix: "Correct" if prefix == PARSE_INFO_PREFIX else "INCORRECT"
        ),
        formatter=Hex(8),
        help_type="int",
    ),
    Entry(
        "parse_code",
        enum=ParseCodes,
        formatter=Hex(2),
        help_type=":py:class:`~vc2_data_tables.ParseCodes`",
    ),
    Entry("next_parse_offset", help_type="int"),
    Entry("previous_parse_offset", help_type="int"),
    Entry(
        "_offset",
        help_type="int",
        help="""
            Computed value. The byte offset of the start of this parse_info
            block in the bitstream.
        """,
    ),
    help="""
        (10.5.1) Parse info header defined by ``parse_info()``.
    """,
)

vc2_default_values[ParseInfo] = ParseInfo(
    padding=bitarray(),
    parse_info_prefix=PARSE_INFO_PREFIX,
    parse_code=ParseCodes.end_of_sequence,
    next_parse_offset=0,
    previous_parse_offset=0,
)

################################################################################
# sequence_header header and internal structures
################################################################################

ParseParameters = fixeddict(
    "ParseParameters",
    Entry("major_version", help_type="int"),
    Entry("minor_version", help_type="int"),
    Entry("profile", enum=Profiles, help_type=":py:class:`~vc2_data_tables.Profiles`"),
    Entry("level", enum=Levels, help_type=":py:class:`~vc2_data_tables.Levels`"),
    help="""
        (11.2.1) Sequence header defined by ``parse_parameters()``.
    """,
)

vc2_default_values[ParseParameters] = ParseParameters(
    major_version=3,
    minor_version=0,
    profile=Profiles.high_quality,
    level=Levels.unconstrained,
)

FrameSize = fixeddict(
    "FrameSize",
    Entry("custom_dimensions_flag", formatter=Bool(), help_type="bool"),
    Entry("frame_width", help_type="int"),
    Entry("frame_height", help_type="int"),
    help="""
        (11.4.3) Frame size override defined by ``frame_size()``.
    """,
)

vc2_default_values[FrameSize] = FrameSize(
    custom_dimensions_flag=False,
    frame_width=1,
    frame_height=1,
)

ColorDiffSamplingFormat = fixeddict(
    "ColorDiffSamplingFormat",
    Entry("custom_color_diff_format_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "color_diff_format_index",
        enum=ColorDifferenceSamplingFormats,
        help_type=":py:class:`~vc2_data_tables.ColorDifferenceSamplingFormats`",
    ),
    help="""
        (11.4.4) Color-difference sampling override defined by
        ``color_diff_sampling_format()``.
    """,
)

vc2_default_values[ColorDiffSamplingFormat] = ColorDiffSamplingFormat(
    custom_color_diff_format_flag=False,
    color_diff_format_index=ColorDifferenceSamplingFormats.color_4_4_4,
)

ScanFormat = fixeddict(
    "ScanFormat",
    Entry("custom_scan_format_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "source_sampling",
        enum=SourceSamplingModes,
        help_type=":py:class:`~vc2_data_tables.SourceSamplingModes`",
    ),
    help="""
        (11.4.5) Scan format override defined by ``scan_format()``.
    """,
)

vc2_default_values[ScanFormat] = ScanFormat(
    custom_scan_format_flag=False,
    source_sampling=SourceSamplingModes.progressive,
)

FrameRate = fixeddict(
    "FrameRate",
    Entry("custom_frame_rate_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetFrameRates,
        help_type=":py:class:`~vc2_data_tables.PresetFrameRates`",
    ),
    Entry("frame_rate_numer", help_type="int"),
    Entry("frame_rate_denom", help_type="int"),
    help="""
        (11.4.6) Frame-rate override defined by ``frame_rate()``.
    """,
)

vc2_default_values[FrameRate] = FrameRate(
    custom_frame_rate_flag=False,
    index=PresetFrameRates.fps_25,
    frame_rate_numer=25,
    frame_rate_denom=1,
)

PixelAspectRatio = fixeddict(
    "PixelAspectRatio",
    Entry("custom_pixel_aspect_ratio_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetPixelAspectRatios,
        help_type=":py:class:`~vc2_data_tables.PresetPixelAspectRatios`",
    ),
    Entry("pixel_aspect_ratio_numer", help_type="int"),
    Entry("pixel_aspect_ratio_denom", help_type="int"),
    help="""
        (11.4.7) Pixel aspect ratio override defined by
        ``pixel_aspect_ratio()`` (errata: also listed as ``aspect_ratio()`` in
        some parts of the spec).
    """,
)

vc2_default_values[PixelAspectRatio] = PixelAspectRatio(
    custom_pixel_aspect_ratio_flag=False,
    index=PresetPixelAspectRatios.ratio_1_1,
    pixel_aspect_ratio_numer=1,
    pixel_aspect_ratio_denom=1,
)

CleanArea = fixeddict(
    "CleanArea",
    Entry("custom_clean_area_flag", formatter=Bool(), help_type="bool"),
    Entry("clean_width", help_type="int"),
    Entry("clean_height", help_type="int"),
    Entry("left_offset", help_type="int"),
    Entry("top_offset", help_type="int"),
    help="""
        (11.4.8) Clean areas override defined by ``clean_area()``.
    """,
)

vc2_default_values[CleanArea] = CleanArea(
    custom_clean_area_flag=False,
    clean_width=1,
    clean_height=1,
    left_offset=0,
    top_offset=0,
)

SignalRange = fixeddict(
    "SignalRange",
    Entry("custom_signal_range_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetSignalRanges,
        help_type=":py:class:`~vc2_data_tables.PresetSignalRanges`",
    ),
    Entry("luma_offset", help_type="int"),
    Entry("luma_excursion", help_type="int"),
    Entry("color_diff_offset", help_type="int"),
    Entry("color_diff_excursion", help_type="int"),
    help="""
        (11.4.9) Signal range override defined by ``signal_range()``.
    """,
)

vc2_default_values[SignalRange] = SignalRange(
    custom_signal_range_flag=False,
    index=PresetSignalRanges.video_8bit_full_range,
    luma_offset=0,
    luma_excursion=1,
    color_diff_offset=0,
    color_diff_excursion=1,
)

ColorPrimaries = fixeddict(
    "ColorPrimaries",
    Entry("custom_color_primaries_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetColorPrimaries,
        help_type=":py:class:`~vc2_data_tables.PresetColorPrimaries`",
    ),
    help="""
        (11.4.10.2) Color primaries override defined by ``color_primaries()``.
    """,
)

vc2_default_values[ColorPrimaries] = ColorPrimaries(
    custom_color_primaries_flag=False,
    index=PresetColorPrimaries.hdtv,
)


ColorMatrix = fixeddict(
    "ColorMatrix",
    Entry("custom_color_matrix_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetColorMatrices,
        help_type=":py:class:`~vc2_data_tables.PresetColorMatrices`",
    ),
    help="""
        (11.4.10.3) Color matrix override defined by ``color_matrix()``.
    """,
)

vc2_default_values[ColorMatrix] = ColorMatrix(
    custom_color_matrix_flag=False,
    index=PresetColorMatrices.hdtv,
)

TransferFunction = fixeddict(
    "TransferFunction",
    Entry("custom_transfer_function_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetTransferFunctions,
        help_type=":py:class:`~vc2_data_tables.PresetTransferFunctions`",
    ),
    help="""
        (11.4.10.4) Transfer function override defined by
        ``transfer_function()``.
    """,
)

vc2_default_values[TransferFunction] = TransferFunction(
    custom_transfer_function_flag=False,
    index=PresetTransferFunctions.tv_gamma,
)

ColorSpec = fixeddict(
    "ColorSpec",
    Entry("custom_color_spec_flag", formatter=Bool(), help_type="bool"),
    Entry(
        "index",
        enum=PresetColorSpecs,
        help_type=":py:class:`~vc2_data_tables.PresetColorSpecs`",
    ),
    Entry("color_primaries", help_type=":py:class:`ColorPrimaries`"),
    Entry("color_matrix", help_type=":py:class:`ColorMatrix`"),
    Entry("transfer_function", help_type=":py:class:`TransferFunction`"),
    help="""
        (11.4.10.1) Color specification override defined by ``color_spec()``.
    """,
)

vc2_default_values[ColorSpec] = ColorSpec(
    custom_color_spec_flag=False,
    index=PresetColorSpecs.hdtv,
)

vc2_fixeddict_nesting[ColorSpec] = [ColorPrimaries, ColorMatrix, TransferFunction]

SourceParameters = fixeddict(
    "SourceParameters",
    Entry("frame_size", help_type=":py:class:`FrameSize`"),
    Entry(
        "color_diff_sampling_format", help_type=":py:class:`ColorDiffSamplingFormat`"
    ),
    Entry("scan_format", help_type=":py:class:`ScanFormat`"),
    Entry("frame_rate", help_type=":py:class:`FrameRate`"),
    Entry("pixel_aspect_ratio", help_type=":py:class:`PixelAspectRatio`"),
    Entry("clean_area", help_type=":py:class:`CleanArea`"),
    Entry("signal_range", help_type=":py:class:`SignalRange`"),
    Entry("color_spec", help_type=":py:class:`ColorSpec`"),
    help="""
        (11.4.1) Video format overrides defined by ``source_parameters()``.
    """,
)

vc2_default_values[SourceParameters] = SourceParameters()

vc2_fixeddict_nesting[SourceParameters] = [
    FrameSize,
    ColorDiffSamplingFormat,
    ScanFormat,
    FrameRate,
    PixelAspectRatio,
    CleanArea,
    SignalRange,
    ColorSpec,
]

SequenceHeader = fixeddict(
    "SequenceHeader",
    Entry("parse_parameters", help_type=":py:class:`ParseParameters`"),
    Entry(
        "base_video_format",
        enum=BaseVideoFormats,
        help_type=":py:class:`~vc2_data_tables.BaseVideoFormats`",
    ),
    Entry("video_parameters", help_type=":py:class:`SourceParameters`"),
    Entry(
        "picture_coding_mode",
        enum=PictureCodingModes,
        help_type=":py:class:`~vc2_data_tables.PictureCodingModes`",
    ),
    help="""
        (11.1) Sequence header defined by ``sequence_header()``.
    """,
)

vc2_default_values[SequenceHeader] = SequenceHeader(
    base_video_format=BaseVideoFormats.custom_format,
    picture_coding_mode=PictureCodingModes.pictures_are_frames,
)

vc2_fixeddict_nesting[SequenceHeader] = [ParseParameters, SourceParameters]


################################################################################
# auxiliary_data and padding
################################################################################

AuxiliaryData = fixeddict(
    "AuxiliaryData",
    Entry("bytes", formatter=Bytes(), help_type="bytes"),
    help="""
        (10.4.4) Auxiliary data block (as per auxiliary_data()).
    """,
)

vc2_default_values[AuxiliaryData] = AuxiliaryData(
    bytes=b"",
)


Padding = fixeddict(
    "Padding",
    Entry("bytes", formatter=Bytes(), help_type="bytes"),
    help="""
        (10.4.5) Padding data block (as per padding()).
    """,
)

vc2_default_values[Padding] = Padding(
    bytes=b"",
)


################################################################################
# transform_parameters and associated structures
################################################################################

ExtendedTransformParameters = fixeddict(
    "ExtendedTransformParameters",
    Entry("asym_transform_index_flag", help_type="bool"),
    Entry(
        "wavelet_index_ho",
        enum=WaveletFilters,
        help_type=":py:class:`~vc2_data_tables.WaveletFilters`",
    ),
    Entry("asym_transform_flag", help_type="bool"),
    Entry("dwt_depth_ho", help_type="int"),
    help="""
        (12.4.4.1) Extended (horizontal-only) wavelet transform parameters
        defined by ``extended_transform_parameters()``.
    """,
)

vc2_default_values[ExtendedTransformParameters] = ExtendedTransformParameters(
    asym_transform_index_flag=False,
    wavelet_index_ho=WaveletFilters.haar_with_shift,
    asym_transform_flag=False,
    dwt_depth_ho=0,
)

SliceParameters = fixeddict(
    "SliceParameters",
    Entry("slices_x", help_type="int"),
    Entry("slices_y", help_type="int"),
    Entry("slice_bytes_numerator", help_type="int"),
    Entry("slice_bytes_denominator", help_type="int"),
    Entry("slice_prefix_bytes", help_type="int"),
    Entry("slice_size_scaler", help_type="int"),
    help="""
        (12.4.5.2) Slice dimension parameters defined by
        ``slice_parameters()``.
    """,
)

vc2_default_values[SliceParameters] = SliceParameters(
    slices_x=1,
    slices_y=1,
    slice_bytes_numerator=1,
    slice_bytes_denominator=1,
    slice_prefix_bytes=0,
    slice_size_scaler=1,
)

QuantMatrix = fixeddict(
    "QuantMatrix",
    Entry("custom_quant_matrix", help_type="bool"),
    Entry(
        "quant_matrix",
        help_type="[int, ...]",
        help="Quantization matrix values in bitstream order.",
    ),
    help="""
        (12.4.5.3) Custom quantisation matrix override defined by
        ``quant_matrix()``.
    """,
)

vc2_default_values[QuantMatrix] = QuantMatrix(
    custom_quant_matrix=False,
    quant_matrix=0,
)

TransformParameters = fixeddict(
    "TransformParameters",
    Entry(
        "wavelet_index",
        enum=WaveletFilters,
        help_type=":py:class:`~vc2_data_tables.WaveletFilters`",
    ),
    Entry("dwt_depth", help_type="int"),
    Entry(
        "extended_transform_parameters",
        help_type=":py:class:`ExtendedTransformParameters`",
    ),
    Entry("slice_parameters", help_type=":py:class:`SliceParameters`"),
    Entry("quant_matrix", help_type=":py:class:`QuantMatrix`"),
    help="""
        (12.4.1) Wavelet transform parameters defined by
        ``transform_parameters()``.
    """,
)

vc2_default_values[TransformParameters] = TransformParameters(
    wavelet_index=WaveletFilters.haar_with_shift,
    dwt_depth=0,
)

vc2_fixeddict_nesting[TransformParameters] = [
    ExtendedTransformParameters,
    SliceParameters,
    QuantMatrix,
]

################################################################################
# Slice related structures
#
# Note that in the VC2 spec, slices are unpacked into a series of 2D arrays
# according to the transform to be performed. By contrast, the structures
# defined below mimic the layout of the transform data in the bitstream.
################################################################################

LDSlice = fixeddict(
    "LDSlice",
    Entry("qindex", help_type="int"),
    Entry("slice_y_length", help_type="int"),
    Entry(
        "y_transform",
        formatter=List(),
        help_type="[int, ...]",
        help="""
            Slice luma transform coefficients in bitstream order.
        """,
    ),
    Entry(
        "c_transform",
        formatter=List(),
        help_type="[int, ...]",
        help="""
            Slice interleaved colordifference transform coefficients in
            bitstream order.
        """,
    ),
    Entry(
        "y_block_padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Unused bits from y_transform bounded block.",
    ),
    Entry(
        "c_block_padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Unused bits from c_transform bounded block.",
    ),
    Entry("_sx", help_type="int", help="Computed value. Slice coordinates."),
    Entry("_sy", help_type="int", help="Computed value. Slice coordinates."),
    help="""
        (13.5.3.1) The data associated with a single low-delay slice, defined
        by ``ld_slice()``.
    """,
)

vc2_default_values[LDSlice] = LDSlice(
    qindex=0,
    slice_y_length=0,
    y_transform=0,
    c_transform=0,
    y_block_padding=bitarray(),
    c_block_padding=bitarray(),
)

HQSlice = fixeddict(
    "HQSlice",
    Entry("prefix_bytes", formatter=Bytes(), help_type="bytes"),
    Entry("qindex", help_type="int"),
    Entry("slice_y_length", help_type="int"),
    Entry("slice_c1_length", help_type="int"),
    Entry("slice_c2_length", help_type="int"),
    Entry(
        "y_transform",
        formatter=List(),
        help_type="[int, ...]",
        help="Slice luma transform coefficients in bitstream order.",
    ),
    Entry(
        "c1_transform",
        formatter=List(),
        help_type="[int, ...]",
        help="Slice color difference 1 transform coefficients in bitstream order.",
    ),
    Entry(
        "c2_transform",
        formatter=List(),
        help_type="[int, ...]",
        help="Slice color difference 2 transform coefficients in bitstream order.",
    ),
    Entry(
        "y_block_padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Unused bits in y_transform bounded block",
    ),
    Entry(
        "c1_block_padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Unused bits in c1_transform bounded block",
    ),
    Entry(
        "c2_block_padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Unused bits in c2_transform bounded block",
    ),
    Entry("_sx", help_type="int", help="Computed value. Slice coordinates."),
    Entry("_sy", help_type="int", help="Computed value. Slice coordinates."),
    help="""
        (13.5.4) The data associated with a single high-quality slice, defined
        by ``hq_slice()``.
    """,
)

vc2_default_values[HQSlice] = HQSlice(
    prefix_bytes=b"",
    qindex=0,
    slice_y_length=0,
    slice_c1_length=0,
    slice_c2_length=0,
    y_transform=0,
    c1_transform=0,
    c2_transform=0,
    y_block_padding=bitarray(),
    c1_block_padding=bitarray(),
    c2_block_padding=bitarray(),
)

################################################################################
# picture_parse and associated structures
################################################################################


PictureHeader = fixeddict(
    "PictureHeader",
    Entry("picture_number", help_type="int"),
    help="""
        (12.2) Picture header information defined by ``picture_header()``.
    """,
)

vc2_default_values[PictureHeader] = PictureHeader(
    picture_number=0,
)

TransformData = fixeddict(
    "TransformData",
    Entry(
        "ld_slices",
        formatter=List(formatter=Object()),
        help_type="[:py:class:`LDSlice`, ...]",
    ),
    Entry(
        "hq_slices",
        formatter=List(formatter=Object()),
        help_type="[:py:class:`HQSlice`, ...]",
    ),
    Entry(
        "_state",
        help_type=":py:class:`~vc2_conformance.pseudocode.state.State`",
        help="""
            Computed value. A copy of the
            :py:class:`~vc2_conformance.pseudocode.state.State` dictionary held
            when processing this transform data. May be used to work out how
            the deserialised values correspond to transform components within
            the slices above.
        """,
    ),
    help="""
        (13.5.2) Transform coefficient data slices read by
        ``transform_data()``.
    """,
)

vc2_default_values[TransformData] = TransformData()

vc2_fixeddict_nesting[TransformData] = [LDSlice, HQSlice]

WaveletTransform = fixeddict(
    "WaveletTransform",
    Entry("transform_parameters", help_type=":py:class:`TransformParameters`"),
    Entry(
        "padding",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Byte alignment padding bits.",
    ),
    Entry("transform_data", help_type=":py:class:`TransformData`"),
    help="""
        (12.3) Wavelet parameters and coefficients defined by
        ``wavelet_transform()``.
    """,
)

vc2_default_values[WaveletTransform] = WaveletTransform(
    padding=bitarray(),
)

vc2_fixeddict_nesting[WaveletTransform] = [TransformParameters, TransformData]

PictureParse = fixeddict(
    "PictureParse",
    Entry(
        "padding1",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Picture header byte alignment padding bits.",
    ),
    Entry("picture_header", help_type=":py:class:`PictureHeader`"),
    Entry(
        "padding2",
        formatter=Bits(),
        help_type=":py:class:`~bitarray.bitarray`",
        help="Wavelet transform byte alignment padding bits.",
    ),
    Entry("wavelet_transform", help_type=":py:class:`WaveletTransform`"),
    help="""
        (12.1) A picture data unit defined by ``picture_parse()``
    """,
)

vc2_default_values[PictureParse] = PictureParse(
    padding1=bitarray(),
    padding2=bitarray(),
)

vc2_fixeddict_nesting[PictureParse] = [PictureHeader, WaveletTransform]


################################################################################
# fragment_parse and associated structures
################################################################################

FragmentHeader = fixeddict(
    "FragmentHeader",
    Entry("picture_number", help_type="int"),
    Entry("fragment_data_length", help_type="int"),
    Entry("fragment_slice_count", help_type="int"),
    Entry("fragment_x_offset", help_type="int"),
    Entry("fragment_y_offset", help_type="int"),
    help="""
        (14.2) Fragment header defined by ``fragment_header()``.
    """,
)

vc2_default_values[FragmentHeader] = FragmentHeader(
    picture_number=0,
    fragment_data_length=0,
    fragment_slice_count=0,
    fragment_x_offset=0,
    fragment_y_offset=0,
)

FragmentData = fixeddict(
    "FragmentData",
    Entry(
        "ld_slices",
        formatter=List(formatter=Object()),
        help_type="[:py:class:`LDSlice`, ...]",
    ),
    Entry(
        "hq_slices",
        formatter=List(formatter=Object()),
        help_type="[:py:class:`HQSlice`, ...]",
    ),
    Entry(
        "_state",
        help_type=":py:class:`~vc2_conformance.pseudocode.state.State`",
        help="""
            Computed value. A copy of the
            :py:class:`~vc2_conformance.pseudocode.state.State` dictionary held
            when processing this fragment data. May be used to work out how the
            deserialised values correspond to transform components within the
            slices above.
        """,
    ),
    help="""
        (14.4) Transform coefficient data slices read by ``fragment_data()``.
    """,
)

vc2_default_values[FragmentData] = FragmentData()

vc2_fixeddict_nesting[FragmentData] = [LDSlice, HQSlice]

FragmentParse = fixeddict(
    "FragmentParse",
    Entry("fragment_header", help_type=":py:class:`FragmentHeader`"),
    Entry("transform_parameters", help_type=":py:class:`TransformParameters`"),
    Entry("fragment_data", help_type=":py:class:`FragmentData`"),
    help="""
        (14.1) A fragment data unit defined by ``fragment_parse()`` containing
        part of a picture.
    """,
)

vc2_default_values[FragmentParse] = FragmentParse()

vc2_fixeddict_nesting[FragmentParse] = [
    FragmentHeader,
    TransformParameters,
    FragmentData,
]

################################################################################
# Sequences
################################################################################

DataUnit = fixeddict(
    "DataUnit",
    Entry("parse_info", help_type=":py:class:`ParseInfo`"),
    Entry("sequence_header", help_type=":py:class:`SequenceHeader`"),
    Entry("picture_parse", help_type=":py:class:`PictureParse`"),
    Entry("fragment_parse", help_type=":py:class:`FragmentParse`"),
    Entry("auxiliary_data", help_type=":py:class:`AuxiliaryData`"),
    Entry("padding", help_type=":py:class:`Padding`"),
    help="""
        A data unit (e.g. sequence header or picture) and its associated parse
        info.  Based on the values read by parse_sequence() (10.4.1) in each
        iteration.
    """,
)

vc2_default_values[DataUnit] = DataUnit()

vc2_fixeddict_nesting[DataUnit] = [
    ParseInfo,
    SequenceHeader,
    PictureParse,
    FragmentParse,
    AuxiliaryData,
    Padding,
]

Sequence = fixeddict(
    "Sequence",
    Entry(
        "data_units",
        formatter=MultilineList(heading=""),
        help_type="[:py:class:`DataUnit`, ...]",
    ),
    Entry(
        "_state",
        help_type=":py:class:`~vc2_conformance.pseudocode.state.State`",
        help="""
            Computed value. The
            :py:class:`~vc2_conformance.pseudocode.state.State` object being
            populated by the parser.
        """,
    ),
    help="""
        (10.4.1) A VC-2 sequence.
    """,
)

vc2_default_values[Sequence] = Sequence()

vc2_fixeddict_nesting[Sequence] = [DataUnit]

Stream = fixeddict(
    "Stream",
    Entry(
        "sequences",
        formatter=MultilineList(heading=""),
        help_type="[:py:class:`Sequence`, ...]",
    ),
    help="""
        (10.3) A VC-2 stream.
    """,
)

vc2_default_values[Stream] = Stream()

vc2_fixeddict_nesting[Stream] = [Sequence]



"""
Mathematical functions and operators defined in the spec (5.5.3)
"""

import math

from vc2_conformance.pseudocode.metadata import ref_pseudocode

__all__ = [
    "intlog2_float",
    "intlog2",
    "sign",
    "clip",
    "mean",
]


@ref_pseudocode(name="intlog2", deviation="inferred_implementation")
def intlog2_float(n):
    """
    (5.5.3) Implemented as described in the spec, requiring floating point
    arithmetic.

    In practice, the alternative implementation in :py:func:`intlog2` should be
    used instead since it does not rely on floating point arithmetic (and is
    therefore faster and has unlimited precision).
    """
    return int(math.ceil(math.log(n, 2)))


@ref_pseudocode(deviation="alternative_implementation")
def intlog2(n):
    """
    (5.5.3) Implemented via pure integer, arbitrary-precision operations.
    """
    return (n - 1).bit_length()


@ref_pseudocode
def sign(a):
    """(5.5.3)"""
    if a > 0:
        return 1
    elif a == 0:
        return 0
    elif a < 0:
        return -1


@ref_pseudocode
def clip(a, b, t):
    """(5.5.3)"""
    return min(max(a, b), t)


@ref_pseudocode(deviation="inferred_implementation")
def mean(*S):
    """(5.5.3)"""
    n = len(S)
    return (sum(S) + (n // 2)) // n



"""
Parse code classification functions (Table 10.2)

These functions all take a 'state' dictionary containing at least
``parse_code`` which should be an int or :py:class:`ParseCodes` enum value.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

__all__ = [
    "is_seq_header",
    "is_end_of_sequence",
    "is_auxiliary_data",
    "is_padding_data",
    "is_ld",
    "is_hq",
    "is_picture",
    "is_fragment",
    "using_dc_prediction",
]


@ref_pseudocode
def is_seq_header(state):
    """(Table 10.2)"""
    return state["parse_code"] == 0x00


@ref_pseudocode
def is_end_of_sequence(state):
    """(Table 10.2)"""
    return state["parse_code"] == 0x10


@ref_pseudocode
def is_auxiliary_data(state):
    """(Table 10.2)"""
    return (state["parse_code"] & 0xF8) == 0x20


@ref_pseudocode
def is_padding_data(state):
    """(Table 10.2)"""
    return state["parse_code"] == 0x30


def is_ld(state):
    """(Table 10.2)"""
    return (state["parse_code"] & 0xF8) == 0xC8


def is_hq(state):
    """(Table 10.2)"""
    return (state["parse_code"] & 0xF8) == 0xE8


@ref_pseudocode
def is_picture(state):
    """(Table 10.2)"""
    return (state["parse_code"] & 0x8C) == 0x88


@ref_pseudocode
def is_fragment(state):
    """(Table 10.2)"""
    return (state["parse_code"] & 0x0C) == 0x0C


@ref_pseudocode
def using_dc_prediction(state):
    """(Table 10.2)"""
    return (state["parse_code"] & 0x28) == 0x08



"""
The :py:mod:`vc2_conformance.io` module implements I/O functions for reading
bitstreams from file-like objects. The exposed functions implement the
interface specified in annex (A).


Initialisation
--------------

The :py:func:`init_io` function must be used to initialise a
:py:class:`~vc2_conformance.pseudocode.state.State` dictionary so that it is
ready to read a bitstream.

.. autofunction:: init_io


Determining stream position
---------------------------

The :py:func:`tell` function (below) is used by verification logic to report
and check offsets of values within a bitstream. For example, it may be used to
check ``next_parse_offset`` fields are correct (see
:py:func:`vc2_conformance.decoder.stream.parse_info`).

.. autofunction:: tell


Bitstream recording
-------------------

The VC-2 specification sometimes requires that the coded bitstream
representation of a particular set of repeated fields is consistent within a
bitstream (e.g. see
:py:func:`vc2_conformance.decoder.sequence_header.sequence_header`). To
facilitate this test, the :py:func:`record_bitstream_start` and
:py:func:`record_bitstream_finish` functions may be used to capture the
bitstream bytes read within part of the bitstream.

.. autofunction:: record_bitstream_start

.. autofunction:: record_bitstream_finish


"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.decoder.exceptions import UnexpectedEndOfStream

__all__ = [
    "init_io",
    "record_bitstream_start",
    "record_bitstream_finish",
    "tell",
    "read_byte",
    "is_end_of_stream",
    "read_bit",
    "byte_align",
    "read_bool",
    "read_nbits",
    "read_uint_lit",
    "read_bitb",
    "read_boolb",
    "flush_inputb",
    "read_uint",
    "read_uintb",
    "read_sint",
    "read_sintb",
]


@ref_pseudocode(deviation="inferred_implementation")
def init_io(state, f):
    """
    (A.2.1) Initialise the I/O-related variables in state.

    This function should be called exactly once to initialise the I/O-related
    parts of the state dictionary to their initial state as specified by
    (A.2.1):

        ... a decoder is deemed to maintain a copy of the current byte,
        state[current_byte], and an index to the next bit (in the byte) to be
        read, state[next_bit] ...

    As well as initialising the state["current_byte"] and state["next_bit"]
    fields, this sets the (out-of-spec) state["_file"] entry to the provided
    file-like object.

    Parameters
    ----------
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        The state dictionary to be initialised.
    f : file-like object
        The file to read the bitstream from.
    """
    state["_file"] = f
    read_byte(state)


def record_bitstream_start(state):
    """
    Not part of spec; used for verifying that repeated sequence_headers are
    byte-for-byte identical (11.1).

    This function causes all future bytes read from the bitstream to be logged
    into state["_read_bytes"] until :py:func:`record_bitstream_finish` is
    called.

    Recordings must start byte aligned.
    """
    assert state["next_bit"] == 7, "Recordings must always be byte aligned"
    assert state.get("_recorded_bytes") is None, "Cannot nest recordings"
    state["_recorded_bytes"] = bytearray()


def record_bitstream_finish(state):
    """
    See :py:func:`record_bitstream_start`.

    Returns
    =======
    bytearray
        The bytes read since :py:func:`record_bitstream_start` was called. Any
        unread bits of the final byte will be set to zero.
    """
    recorded_bytes = state["_recorded_bytes"]
    del state["_recorded_bytes"]

    # Add whatever has been used of the current byte
    if state["next_bit"] != 7:
        recorded_bytes.append(
            state["current_byte"] & ~((1 << (state["next_bit"] + 1)) - 1)
        )

    return recorded_bytes


def tell(state):
    """
    Not part of spec; used to log bit offsets in the bitstream.

    Return a (byte, bit) tuple giving the offset of the next bit to be read in
    the stream.
    """
    return (
        state["_file"].tell() - (1 if state["current_byte"] is not None else 0),
        state["next_bit"],
    )


@ref_pseudocode(deviation="inferred_implementation")
def read_byte(state):
    """
    (A.2.2) Load the next byte in the stream into the global state, ready for
    the bits to be read.

    By convention, when the end-of-file is reached, ``state["current_byte"]``
    is set to ``None``. This condition must be checked before every read
    operation.
    """
    # Record the now used-up byte (see record_bitstream_start)
    if "_recorded_bytes" in state:
        state["_recorded_bytes"].append(state["current_byte"])

    # Step 1.
    state["next_bit"] = 7

    # Step 2.
    byte = state["_file"].read(1)
    if len(byte) == 1:
        state["current_byte"] = bytearray(byte)[0]  # Convert byte to int
    else:
        # End of file
        state["current_byte"] = None


@ref_pseudocode(deviation="inferred_implementation")
def is_end_of_stream(state):
    """
    (A.2.5) Determine if we have reached the end of the stream (i.e. the end of
    the file).
    """
    return state["current_byte"] is None


@ref_pseudocode
def read_bit(state):
    """(A.2.3)"""
    ## Begin not in spec
    # From (A.2.1):
    #
    #     ... a decoder is deemed  to maintain a copy of the current byte,
    #     state[current_byte] ...
    #
    # As such, if there is no 'current_byte' when we come to read, we've
    # reached a situation not defined by the specification.
    if state["current_byte"] is None:
        raise UnexpectedEndOfStream()
    ## End not in spec

    bit = (state["current_byte"] >> state["next_bit"]) & 1
    state["next_bit"] -= 1
    if state["next_bit"] < 0:
        state["next_bit"] = 7
        read_byte(state)
    return bit


@ref_pseudocode
def byte_align(state):
    """(A.2.4)"""
    # NB: No check for end-of-stream is required here. If we're at the end of
    # the stream, we'll be on a byte boundary so this is a no-op. If we're not
    # at the end of the stream, we can always read until the end of the current
    # byte, even if it is the last in the stream.
    if state["next_bit"] != 7:
        read_byte(state)


@ref_pseudocode
def read_bool(state):
    """(A.3.2)"""
    if read_bit(state) == 1:
        return True
    else:
        return False


@ref_pseudocode
def read_nbits(state, n):
    """(A.3.3)"""
    val = 0
    for i in range(n):
        val <<= 1
        val += read_bit(state)
    return val


@ref_pseudocode
def read_uint_lit(state, n):
    """(A.3.4)"""
    return read_nbits(state, 8 * n)


@ref_pseudocode
def read_bitb(state):
    """(A.4.2)"""
    if state["bits_left"] == 0:
        return 1
    else:
        state["bits_left"] -= 1
        return read_bit(state)


@ref_pseudocode
def read_boolb(state):
    """(A.4.2)"""
    if read_bitb(state) == 1:
        return True
    else:
        return False


@ref_pseudocode
def flush_inputb(state):
    """(A.4.2)"""
    while state["bits_left"] > 0:
        read_bit(state)
        state["bits_left"] -= 1


@ref_pseudocode
def read_uint(state):
    """(A.4.3)"""
    value = 1
    while read_bit(state) == 0:
        value <<= 1
        if read_bit(state) == 1:
            value += 1
    value -= 1
    return value


@ref_pseudocode
def read_uintb(state):
    """(A.4.3)"""
    value = 1
    while read_bitb(state) == 0:
        value <<= 1
        if read_bitb(state):
            value += 1
    value -= 1
    return value


@ref_pseudocode
def read_sint(state):
    """(A.4.4)"""
    value = read_uint(state)
    if value != 0:
        if read_bit(state) == 1:
            value = -value
    return value


@ref_pseudocode
def read_sintb(state):
    """(A.4.4)"""
    value = read_uintb(state)
    if value != 0:
        if read_bitb(state) == 1:
            value = -value
    return value



"""
The :py:mod:`vc2_conformance.decoder` module contains the components of a VC-2
decoder and bitstream validator. Along with the basic decoding logic,
additional tests are included to check all conditions imposed on bitstreams by
the VC-2 specification.

Usage
-----

The bitstream decoder/validator is exposed to end-users via the
:ref:`vc2-bitstream-validator` command line utility.

This module may also be used directly. The following snippet illustrates how a
VC-2 bitstream might be decoded and verified using this module::

    >>> from vc2_conformance.string_utils import wrap_paragraphs
    >>> from vc2_conformance.pseudocode import State
    >>> from vc2_conformance.decoder import init_io, parse_stream, ConformanceError

    >>> # Create a callback to be called with picture data whenever a picture
    >>> # is decoded from the bitstream.
    >>> def output_picture_callback(picture, video_parameters, picture_coding_mode):
    >>>     print("A picture was decoded...")

    >>> # Create an initial state object ready to read the bitstream
    >>> state = State(_output_picture_callback=output_picture_callback)
    >>> f = open("path/to/bitstream.vc2", "rb")
    >>> init_io(state, f)

    >>> # Decode and validate!
    >>> try:
    ...     parse_stream(state)
    ...     print("Bitstream is valid!")
    ... except ConformanceError as e:
    ...     print("Bitstream is NOT valid:")
    ...     print(wrap_paragraphs(e.explain(), 80))
    Bitstream is NOT valid:
    An invalid parse code, 0x0A, was provided to a parse info header (10.5.1).
    <BLANKLINE>
    See (Table 10.1) for the list of allowed parse codes.
    <BLANKLINE>
    Perhaps this bitstream conforms to an earlier or later version of the VC-2
    standard?


Overview
--------

This decoder is based on the pseudocode published in the VC-2 specification and
consequently follows the same structure as the pseudocode with the
:py:func:`~vc2_conformance.decoder.stream.parse_stream` function being used to
decode a complete stream.

The pseudocode is automatically verified for consistency with the VC-2
specification by the conformance software test suite. Verified pseudocode
functions are annotated with the
:py:func:`~vc2_conformance.pseudocode.metadata.ref_pseudocode` decorator. See
:py:mod:`verification` (in the ``tests/`` directory) for details on the
automated verification process. All bitstream validation logic, which doesn't
form part of the specified pseudocode, appears between ``## Begin not in spec``
and ``## End not in spec`` comments.

All global state is passed around via the
:py:class:`~vc2_conformance.pseudocode.state.State` dictionary. This dictionary
is augmented with a number of additional entries not included in the VC-2
specification but which are necessary for a 'real' decoder implementation (e.g.
an input file handle) and for validation purposes (e.g. recorded offsets of
previous data units). See :py:class:`vc2_conformance.pseudocode.state.State`
for a complete enumeration of these.

Underlying I/O operations are not specified by the VC-2 specification. This
decoder reads streams from file-like objects. See
the :py:mod:`vc2_conformance.decoder.io` module for details. As illustrated in
the example above, the :py:func:`~vc2_conformance.decoder.io.init_io` function
is used to specify the file-like object the stream will be read from.

When conformance errors are detected,
:py:exc:`~vc2_conformance.decoder.exceptions.ConformanceError` exceptions are
thrown. These exceptions provide in-depth human readable explanations of
conformance issues along with suggested invocations of the
:ref:`vc2-bitstream-viewer` tool for diagnosing issues. See the
:py:mod:`vc2_conformance.decoder.exceptions` module for details. These
exceptions are largely thrown directly by validation code spliced into the
pseudocode routines. Some common checks are factored out into their own
'assertions' in the :py:mod:`vc2_conformance.decoder.assertions`.

The decoder logic is organised in line with the sections of the VC-2
specification:

* :py:mod:`vc2_conformance.decoder.io`: (A) Bitstream I/O
* :py:mod:`vc2_conformance.decoder.stream`: (10) Stream syntax
* :py:mod:`vc2_conformance.decoder.sequence_header`: (11) Sequence header
* :py:mod:`vc2_conformance.decoder.picture_syntax`: (12) Picture syntax
* :py:mod:`vc2_conformance.decoder.transform_data_syntax`: (13) Transform data
  syntax
* :py:mod:`vc2_conformance.decoder.fragment_syntax`: (14) Fragment syntax

The :py:mod:`vc2_conformance.decoder` module only includes pseudocode functions
which define additional behaviour not defined by the spec.  Specifically, this
includes performing I/O or additional checks for bitstream validation purposes.
All other pseudocode routines are used 'verbatim' and can be found in
:py:mod:`vc2_conformance.pseudocode`.


Stream I/O
----------

.. automodule:: vc2_conformance.decoder.io


Conformance exceptions
----------------------

.. automodule:: vc2_conformance.decoder.exceptions


Sequence composition restrictions
---------------------------------

Various restrictions may be imposed on choice and order of data unit types in a
sequence. For example, all sequences must start with a sequence header and end
with an end of sequence. Some levels impose additional restrictions such as
prohibiting the mixing of fragments and pictures or requiring sequence headers
to be interleaved between every picture.

Rather than using ad-hoc logic to enforce these restrictions, regular
expressions are used to check the pattern of data unit types. See the
:py:mod:`vc2_conformance.symbol_re` module for details on the regular
expression matching system.


Level constraints
-----------------

VC-2's levels impose additional constraints on bitstreams, for example
restricting some fields to particular ranges of values. Rather than including
ad-hoc validation logic for each level, a 'constraints table' is used.
Bitstream values which may be constrained are checked using
:py:func:`vc2_conformance.decoder.assertions.assert_level_constraint`.

See the :py:mod:`vc2_conformance.constraint_table` module for an introduction
to constraint tables. See the :py:mod:`vc2_conformance.level_constraints`
module for level-related constraint data, including documentation on the
entries included in the levels constraints table.
"""

from vc2_conformance.decoder.exceptions import *

# (A) Bitstream I/O operations
from vc2_conformance.decoder.io import *

# (10) Stream syntax
from vc2_conformance.decoder.stream import *

# (11) Sequence Header
from vc2_conformance.decoder.sequence_header import *

# (12) Picture Syntax
from vc2_conformance.decoder.picture_syntax import *

# (13) Transform Data Syntax
from vc2_conformance.decoder.transform_data_syntax import *

# (14) Fragment Syntax
from vc2_conformance.decoder.fragment_syntax import *



"""
This module is the inverse of the
:py:mod:`vc2_conformance.pseudocode.picture_decoding` module and contains
functions for performing wavelet analysis filtering.

This functionality is not specified by the standard but is used to generate
simple bitstreams (and test cases) in this software (and its test suite).
"""

from vc2_data_tables import LIFTING_FILTERS, LiftingFilterTypes

from vc2_conformance.pseudocode.arrays import (
    new_array,
    width,
    height,
    row,
    column,
)

from vc2_conformance.pseudocode.slice_sizes import (
    subband_width,
    subband_height,
)

from vc2_conformance.pseudocode.picture_decoding import (
    filter_bit_shift,
    SYNTHESIS_LIFTING_FUNCTION_TYPES,
)


__all__ = [
    "picture_encode",
    "forward_wavelet_transform",
    "dwt",
    "h_analysis",
    "vh_analysis",
    "oned_analysis",
    "ANALYSIS_LIFTING_FUNCTION_TYPES",
    "dwt_pad_addition",
    "remove_offset_picture",
    "remove_offset_component",
]


def picture_encode(state, current_picture):
    """
    Inverse of picture_decode (15.2).

    Parameters
    ==========
    state : :py:class:`vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least:

        * ``luma_width``
        * ``luma_height``
        * ``color_diff_width``
        * ``color_diff_height``
        * ``luma_depth``
        * ``color_diff_depth``
        * ``wavelet_index``
        * ``wavelet_index_ho``
        * ``dwt_depth``
        * ``dwt_depth_ho``

        The following entries will be added/replaced with the encoded picture.

        * ``y_transform``
        * ``c1_transform``
        * ``c3_transform``

    current_picture : {component: [[value, ...], ...], ...}
        The picture to be encoded.
    """
    remove_offset_picture(state, current_picture)
    forward_wavelet_transform(state, current_picture)


def forward_wavelet_transform(state, current_picture):
    """(15.3)"""
    for c in ["Y", "C1", "C2"]:
        dwt_pad_addition(state, current_picture[c], c)

    state["y_transform"] = dwt(state, current_picture["Y"])
    state["c1_transform"] = dwt(state, current_picture["C1"])
    state["c2_transform"] = dwt(state, current_picture["C2"])


################################################################################
# Forward (Analysis) Discrete Wavelet Transform
################################################################################


def dwt(state, picture):
    """
    Discrete Wavelet Transform, inverse of idwt (15.4.1)

    Parameters
    ==========
    state : :py:class:`~vc2_conformance.pseudocode.state.State`
        A state dictionary containing at least the following:

        * ``wavelet_index``
        * ``wavelet_index_ho``
        * ``dwt_depth``
        * ``dwt_depth_ho``

    picture : [[pixel_value, ...], ...]
        The synthesized picture.

    Returns
    =======
    coeff_data : {level: {orientation: [[coeff, ...], ...], ...}, ...}
        The complete (power-of-two dimensioned) transform coefficient data.
    """
    coeff_data = {}

    DC_band = picture
    for n in reversed(
        range(state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1)
    ):
        (LL_data, HL_data, LH_data, HH_data) = vh_analysis(state, DC_band)
        DC_band = LL_data
        coeff_data[n] = {}
        coeff_data[n]["HL"] = HL_data
        coeff_data[n]["LH"] = LH_data
        coeff_data[n]["HH"] = HH_data
    for n in reversed(range(1, state["dwt_depth_ho"] + 1)):
        (L_data, H_data) = h_analysis(state, DC_band)
        DC_band = L_data
        coeff_data[n] = {}
        coeff_data[n]["H"] = H_data
    coeff_data[0] = {}
    if state["dwt_depth_ho"] == 0:
        coeff_data[0]["LL"] = DC_band
    else:
        coeff_data[0]["L"] = DC_band

    return coeff_data


def h_analysis(state, data):
    """
    Horizontal-only analysis, inverse of h_synthesis (15.4.2).

    Returns a tuple (L_data, H_data)
    """
    # Bit shift, if required
    shift = filter_bit_shift(state)
    if shift > 0:
        for y in range(0, height(data)):
            for x in range(0, width(data)):
                data[y][x] = data[y][x] << shift

    # Analysis
    for y in range(0, height(data)):
        oned_analysis(row(data, y), state["wavelet_index_ho"])

    # De-interleave the transform data
    L_data = new_array(height(data), width(data) // 2)
    H_data = new_array(height(data), width(data) // 2)
    for y in range(0, (height(data))):
        for x in range(0, (width(data) // 2)):
            L_data[y][x] = data[y][2 * x]
            H_data[y][x] = data[y][(2 * x) + 1]

    return (L_data, H_data)


def vh_analysis(state, data):
    """
    Interleaved vertical and horizontal analysis, inverse of vh_synthesis (15.4.3).

    Returns a tuple (LL_data, HL_data, LH_data, HH_data)
    """
    # Bit shift, if required
    shift = filter_bit_shift(state)
    if shift > 0:
        for y in range(0, height(data)):
            for x in range(0, width(data)):
                data[y][x] = data[y][x] << shift

    # Analysis
    for y in range(0, height(data)):
        oned_analysis(row(data, y), state["wavelet_index_ho"])
    for x in range(0, width(data)):
        oned_analysis(column(data, x), state["wavelet_index"])

    # De-interleave the transform data
    LL_data = new_array(height(data) // 2, width(data) // 2)
    HL_data = new_array(height(data) // 2, width(data) // 2)
    LH_data = new_array(height(data) // 2, width(data) // 2)
    HH_data = new_array(height(data) // 2, width(data) // 2)
    for y in range(0, (height(data) // 2)):
        for x in range(0, (width(data) // 2)):
            LL_data[y][x] = data[2 * y][2 * x]
            HL_data[y][x] = data[2 * y][2 * x + 1]
            LH_data[y][x] = data[2 * y + 1][2 * x]
            HH_data[y][x] = data[2 * y + 1][2 * x + 1]

    return (LL_data, HL_data, LH_data, HH_data)


def oned_analysis(A, filter_index):
    """
    Inverse of oned_synthesis (15.4.4.1) and (15.4.4.3). Acts in-place on 'A'
    """
    filter_params = LIFTING_FILTERS[filter_index]

    for stage in reversed(filter_params.stages):
        lift_fn = ANALYSIS_LIFTING_FUNCTION_TYPES[stage.lift_type]
        lift_fn(A, stage.L, stage.D, stage.taps, stage.S)


ANALYSIS_LIFTING_FUNCTION_TYPES = {
    new_type: SYNTHESIS_LIFTING_FUNCTION_TYPES[old_type]
    for (new_type, old_type) in [
        # Swap 'adds' for 'subtracts'
        (LiftingFilterTypes.even_add_odd, LiftingFilterTypes.even_subtract_odd),
        (LiftingFilterTypes.even_subtract_odd, LiftingFilterTypes.even_add_odd),
        (LiftingFilterTypes.odd_add_even, LiftingFilterTypes.odd_subtract_even),
        (LiftingFilterTypes.odd_subtract_even, LiftingFilterTypes.odd_add_even),
    ]
}
"""
Lookup from lifting function ID to function implementation for implementing
wavelet analysis.
"""


################################################################################
# Padding addition
################################################################################


def dwt_pad_addition(state, pic, c):
    """
    Inverse of idwt_pad_removal (15.4.5): pads a picture to a size compatible
    with wavelet filtering at the level specified by the provided
    :py:class:`~vc2_conformance.pseudocode.state.State`.

    Extra values are obtained by copying the final pixels in the existing rows
    and columns.
    """
    top_level = state["dwt_depth"] + state["dwt_depth_ho"] + 1
    width = subband_width(state, top_level, c)
    height = subband_height(state, top_level, c)

    # Copy-extend every row
    for pic_row in pic:
        value = pic_row[-1]
        while len(pic_row) < width:
            pic_row.append(value)

    # Copy extend every column
    while len(pic) < height:
        pic.append(pic[-1][:])


################################################################################
# Pixel value offset removal
################################################################################


def remove_offset_picture(state, current_picture):
    """Inverse of offset_picture (15.5). Centers picture values around zero."""
    for c in ["Y", "C1", "C2"]:
        remove_offset_component(state, current_picture[c], c)


def remove_offset_component(state, comp_data, c):
    """
    Inverse of offset_component (15.5). Centers picture values around zero.

    Parameters
    ==========
    state : :py:class:`vc2_conformance.pseudocode.state.State`
        Where ``luma_depth`` and ``color_diff_depth`` are defined.
    current_picture : {comp: [[pixel_value, ...], ...], ...}
        Will be mutated in-place.
    """
    for y in range(0, height(comp_data)):
        for x in range(0, width(comp_data)):
            if c == "Y":
                comp_data[y][x] -= 2 ** (state["luma_depth"] - 1)
            else:
                comp_data[y][x] -= 2 ** (state["color_diff_depth"] - 1)



import logging

from vc2_data_tables import Profiles

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.constraint_table import allowed_values_for

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.codec_features import codec_features_to_trivial_level_constraints

from vc2_conformance.encoder import make_sequence

from vc2_conformance.picture_generators import mid_gray

from vc2_conformance.test_cases.decoder.common import (
    make_dummy_end_of_sequence,
    iter_slices_in_sequence,
)


def iter_slice_parameters_in_sequence(sequence):
    """
    Iterate over all of the slice parameters blocks in a sequence.

    Generates a series of references to
    :py:class:`~vc2_conformance.bitstream.SliceParameters` dicts, one for each
    slice present in the provided
    :py:class:`~vc2_conformance.bitstream.Sequence`.
    """
    for data_unit in sequence["data_units"]:
        if "picture_parse" in data_unit:
            wt = data_unit["picture_parse"]["wavelet_transform"]
            yield wt["transform_parameters"]["slice_parameters"]
        elif "fragment_parse" in data_unit:
            if "transform_parameters" in data_unit["fragment_parse"]:
                yield data_unit["fragment_parse"]["transform_parameters"][
                    "slice_parameters"
                ]


@decoder_test_case_generator
def slice_prefix_bytes(codec_features):
    """
    **Tests the decoder can handle a non-zero number of slice prefix bytes.**

    Produces test cases with a non-zero number of slice prefix bytes
    containing the following values:

    ``slice_prefix_bytes[zeros]``
        All slice prefix bytes are 0x00.

    ``slice_prefix_bytes[ones]``
        All slice prefix bytes are 0xFF.

    ``slice_prefix_bytes[end_of_sequence]``
        All slice prefix bytes contain bits which encode an end of sequence
        data unit (10.4).

    These test cases apply only to the high quality profile and are omitted
    when the low delay profile is used.
    """
    # This test only applies to high quality codecs
    if codec_features["profile"] != Profiles.high_quality:
        return

    constrained_values = codec_features_to_trivial_level_constraints(codec_features)
    allowed_slice_prefix_bytes = allowed_values_for(
        LEVEL_CONSTRAINTS, "slice_prefix_bytes", constrained_values
    )

    mid_gray_pictures = list(
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        )
    )

    test_cases = [
        ("zeros", b"\x00"),
        ("ones", b"\xFF"),
        ("end_of_sequence", make_dummy_end_of_sequence()),
    ]

    for description, filler in test_cases:
        sequence = make_sequence(codec_features, mid_gray_pictures)

        # Determine how many slice prefix bytes we can fit in our slices
        if codec_features["lossless"]:
            # Lossless slices can be as large as we like; assign enough slice
            # bytes for the full set of filler bytes
            slice_prefix_bytes = len(filler)
        else:
            # Find the space assigned for coefficients in the smallest slice in
            # a fixed-bit-rate stream; we'll replace all slice coefficients
            # with slice prefix bytes.
            slice_prefix_bytes = min(
                (
                    hq_slice["slice_y_length"]
                    + hq_slice["slice_c1_length"]
                    + hq_slice["slice_c2_length"]
                )
                * state["slice_size_scaler"]
                for state, sx, sy, hq_slice in iter_slices_in_sequence(
                    codec_features, sequence
                )
            )

        # Check level constraints allow this slice_prefix_bytes
        #
        # NB: This implementation assumes that either the slice_prefix_bytes
        # field is required to be zero or it is free to be any value. This
        # assumption is verified for all existing VC-2 levels in the tests for
        # this module. Should this assumption be violated, more sophisticated
        # behaviour will be required here...
        if slice_prefix_bytes not in allowed_slice_prefix_bytes:
            continue

        if slice_prefix_bytes < len(filler):
            logging.warning(
                (
                    "Codec '%s' has a very small picture_bytes value "
                    "meaning the slice_prefix_bytes[%s] test case might not "
                    "be as useful as intended."
                ),
                codec_features["name"],
                description,
            )

        # Set the number of slice prefix bytes in all slice parameter headers
        for slice_parameters in iter_slice_parameters_in_sequence(sequence):
            assert slice_parameters["slice_prefix_bytes"] == 0
            slice_parameters["slice_prefix_bytes"] = slice_prefix_bytes

        # Add prefix bytes to all slices
        prefix_bytes = (filler * slice_prefix_bytes)[:slice_prefix_bytes]
        for state, sx, sy, hq_slice in iter_slices_in_sequence(
            codec_features, sequence
        ):
            hq_slice["prefix_bytes"] = prefix_bytes

            # Keep overall slice size the same for lossy (constant bit rate)
            # modes
            if not codec_features["lossless"]:
                total_length = (
                    hq_slice["slice_y_length"]
                    + hq_slice["slice_c1_length"]
                    + hq_slice["slice_c2_length"]
                )

                total_length -= slice_prefix_bytes // state["slice_size_scaler"]

                hq_slice["slice_y_length"] = 0
                hq_slice["slice_c1_length"] = 0
                hq_slice["slice_c2_length"] = total_length

        yield TestCase(Stream(sequences=[sequence]), description)



"""
The :py:mod:`vc2_conformance.level_constraints` module contains definitions of
constraints imposed on VC-2 bitstreams by the various VC-2 level
specifications.

Sequence data unit ordering restrictions
----------------------------------------

Levels may restrict the ordering or choice of data unit types within a
bitstream. These restrictions are described using
:py:mod:`~vc2_conformance.symbol_re` regular expressions provided in
:py:data:`LEVEL_SEQUENCE_RESTRICTIONS`.

.. autodata:: LEVEL_SEQUENCE_RESTRICTIONS
    :annotation: = {level: LevelSequenceRestrictions, ...}

.. autoclass:: LevelSequenceRestrictions


Coding parameter restrictions
-----------------------------

Levels impose various restrictions on bitstream parameters and values. These
restrictions are collected into a constraint table (see
:py:mod:`~vc2_conformance.constraint_table`) in :py:data:`LEVEL_CONSTRAINTS`.

.. autodata:: LEVEL_CONSTRAINTS
    :annotation: = <constraint table>

.. autodata:: LEVEL_CONSTRAINT_ANY_VALUES
    :annotation: = {key: ValueSet, ...}

"""

import os

from collections import namedtuple

from vc2_conformance.constraint_table import (
    read_constraints_from_csv,
    ValueSet,
)

from vc2_data_tables.csv_readers import read_lookup_from_csv

from vc2_data_tables import (
    Levels,
    Profiles,
    BaseVideoFormats,
    ColorDifferenceSamplingFormats,
    SourceSamplingModes,
    PresetFrameRates,
    PresetPixelAspectRatios,
    PresetSignalRanges,
    PresetColorSpecs,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
    PictureCodingModes,
    WaveletFilters,
)


__all__ = [
    "LEVEL_SEQUENCE_RESTRICTIONS",
    "LEVEL_CONSTRAINTS",
    "LEVEL_CONSTRAINT_ANY_VALUES",
]


LevelSequenceRestrictions = namedtuple(
    "LevelSequenceRestrictions",
    "sequence_restriction_explanation,sequence_restriction_regex",
)
"""
Restrictions on sequence orderings for a VC-2 level.

Parameters
----------
sequence_restriction_explanation : str
    A human readable explanation of the restriction imposed (informative).
sequence_restriction_regex : str
    A regular expression describing the sequence ordering allowed which can be
    matched using a :py:class:`~vc2_conformance.symbol_re.Matcher`. Each symbol
    is a :py:class:`~vc2_data_tables.ParseCodes` name string.
"""

LEVEL_SEQUENCE_RESTRICTIONS = read_lookup_from_csv(
    os.path.join(os.path.dirname(__file__), "level_sequence_restrictions.csv"),
    Levels,
    LevelSequenceRestrictions,
)
"""
A lookup from :py:class:`Levels` to :py:class:`LevelSequenceRestrictions`
(loaded from ``vc2_conformance/level_sequence_restrictions.csv``) describing the
restrictions on sequences imposed by each VC-2 level.
"""


LEVEL_CONSTRAINTS = read_constraints_from_csv(
    os.path.join(
        os.path.dirname(__file__),
        "level_constraints.csv",
    )
)
"""
A constraint table (see :py:mod:`vc2_conformance.constraint_table`) loaded
from ``vc2_conformance/level_constraints.csv``.

Constraints which apply due to levels. Keys correspond to particular bitstream
values or properties and are enumerated below:

* (11.2.1)
    * ``level``: int (from the :py:class:`Levels` enum)
    * ``profile``: int (from the :py:class:`Profiles` enum)
    * ``major_version``: int
    * ``minor_version``: int
* (11.1)
    * ``base_video_format``: int (from the :py:class:`BaseVideoFormats` enum)
* (11.4.3)
    * ``custom_dimensions_flag``: bool
    * ``frame_width``: int
    * ``frame_height``: int
* (11.4.4)
    * ``custom_color_diff_format_flag``: bool
    * ``color_diff_format_index``: int (from the :py:class:`ColorDifferenceSamplingFormats` enum)
* (11.4.5)
    * ``custom_scan_format_flag``: bool
    * ``source_sampling``: int (from the :py:class:`SourceSamplingModes` enum)
* (11.4.6)
    * ``custom_frame_rate_flag``: bool
    * ``frame_rate_index``: int (from the :py:class:`PresetFrameRates` enum, or 0)
    * ``frame_rate_numer``: int
    * ``frame_rate_denom``: int
* (11.4.7)
    * ``custom_pixel_aspect_ratio_flag``: bool
    * ``pixel_aspect_ratio_index``: int (from the :py:class:`PresetPixelAspectRatios` enum, or 0)
    * ``pixel_aspect_ratio_numer``: int
    * ``pixel_aspect_ratio_denom``: int
* (11.4.8)
    * ``custom_clean_area_flag``: bool
    * ``clean_width``: int
    * ``clean_height``: int
    * ``left_offset``: int
    * ``top_offset``: int
* (11.4.9)
    * ``custom_signal_range_flag``: bool
    * ``custom_signal_range_index``: int (from the :py:class:`PresetSignalRanges` enum, or 0)
    * ``luma_offset``: int
    * ``luma_excursion``: int
    * ``color_diff_offset``: int
    * ``color_diff_excursion``: int
* (11.4.10)
    * ``custom_color_spec_flag``: bool
    * ``color_spec_index``: int (from the :py:class:`PresetColorSpecs` enum)
    * ``custom_color_primaries_flag``: bool
    * ``color_primaries_index``: int (from the :py:class:`PresetColorPrimaries` enum)
    * ``custom_color_matrix_flag``: bool
    * ``color_matrix_index``: int (from the :py:class:`PresetColorMatrices` enum)
    * ``custom_transfer_function_flag``: bool
    * ``transfer_function_index``: int (from the :py:class:`PresetTransferFunctions` enum)
* (11.1)
    * ``picture_coding_mode``: int (from the :py:class:`PictureCodingModes` enum)
* (12.4.1)
    * ``wavelet_index``: int (from the :py:class:`WaveletFilters` enum)
    * ``dwt_depth``: int
* (12.4.4.1)
    * ``asym_transform_index_flag``: bool
    * ``wavelet_index_ho``: int (from the :py:class:`WaveletFilters` enum)
    * ``asym_transform_flag``: bool
    * ``dwt_depth_ho``: int
* (12.4.5.2)
    * ``slices_x``: int (giving the allowed number of slices in the x dimension)
    * ``slices_y``: int (giving the allowed number of slices in the y dimension)
    * ``slices_have_same_dimensions``: bool. True iff all slices contain
      exactly the same number of transform components.
    * ``slice_bytes_numerator``: int
    * ``slice_bytes_denominator``: int
    * ``slice_prefix_bytes``: int
    * ``slice_size_scaler``: int
* (12.4.5.3)
    * ``custom_quant_matrix``: bool
    * ``quant_matrix_values``: int (giving the allowed values within a custom
      quantisation matrix).
* (13.5.3)
    * ``qindex``: int (the allowed qindex values as defined by individual slices)
* (13.5.3.2)
    * ``total_slice_bytes``: int (total number of bytes allowed in a high quality
      picture slice, including all prefix bytes and slice size fields.

See also: :py:data:`LEVEL_CONSTRAINT_ANY_VALUES`.
"""

LEVEL_CONSTRAINT_ANY_VALUES = {
    "level": ValueSet(*Levels),
    "profile": ValueSet(*Profiles),
    "base_video_format": ValueSet(*BaseVideoFormats),
    "custom_dimensions_flag": ValueSet(False, True),
    "custom_color_diff_format_flag": ValueSet(False, True),
    "color_diff_format_index": ValueSet(*ColorDifferenceSamplingFormats),
    "custom_scan_format_flag": ValueSet(False, True),
    "source_sampling": ValueSet(*SourceSamplingModes),
    "custom_frame_rate_flag": ValueSet(False, True),
    "frame_rate_index": ValueSet(0, *PresetFrameRates),
    "custom_pixel_aspect_ratio_flag": ValueSet(False, True),
    "pixel_aspect_ratio_index": ValueSet(0, *PresetPixelAspectRatios),
    "custom_clean_area_flag": ValueSet(False, True),
    "custom_signal_range_flag": ValueSet(False, True),
    "custom_signal_range_index": ValueSet(0, *PresetSignalRanges),
    "custom_color_spec_flag": ValueSet(False, True),
    "color_spec_index": ValueSet(0, *PresetColorSpecs),
    "custom_color_primaries_flag": ValueSet(False, True),
    "color_primaries_index": ValueSet(*PresetColorPrimaries),
    "custom_color_matrix_flag": ValueSet(False, True),
    "color_matrix_index": ValueSet(*PresetColorMatrices),
    "custom_transfer_function_flag": ValueSet(False, True),
    "transfer_function_index": ValueSet(*PresetTransferFunctions),
    "picture_coding_mode": ValueSet(*PictureCodingModes),
    "wavelet_index": ValueSet(*WaveletFilters),
    "asym_transform_index_flag": ValueSet(False, True),
    "wavelet_index_ho": ValueSet(*WaveletFilters),
    "asym_transform_flag": ValueSet(False, True),
    "slices_have_same_dimensions": ValueSet(False, True),
    "custom_quant_matrix": ValueSet(False, True),
}
"""
For keys in :py:data:`LEVEL_CONSTRAINTS` which may hold
:py:class:`~vc2_conformance.constraint_table.AnyValue`, defines an explicit
:py:class:`~vc2_conformance.constraint_table.ValueSet` defining all valid
values, for example when the key refers to an enumerated value. Where the range
of allowed values is truly open ended, no value is provided in this dictionary.
"""
